"""
Examples for differential equations systems with dynamic control.

Here we provide several systems whose state is governed by differential
equations that allow us to plug in a controller. The controller receives
the state of the system as input and its output is added to one (or more) of
the differential equations.

- The :mod:`~moptipyapps.dynamic_control.systems.stuart_landau` system has a
  two-dimensional state space. The goal is to navigate the system into the
  origin. Without control, the system will converge to oscillate in a circle
  around the origin with a diameter of sqrt(0.1).
- The :mod:`~moptipyapps.dynamic_control.systems.lorenz` system has a
  three-dimensional state space. The goal is again to move the system to the
  origin using a control strategy (and to keep it there). Without control, it
  will converge to a double-circle oscillation.
"""



"""
A cubic controller.

A cubic controller is a function where all value of the state vector enter
the computation plainly, squared, and raised to the third power. The powers
of their combinations do not exceed the third power, e.g., the controller is
a linear combination of A, B, A², AB, B², A³, A²B, B²A, and B³ if the state
has values A and B. The controller represents the multipliers for these
coefficients.
"""

from typing import Final

import numba  # type: ignore
import numpy as np

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.system import System


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __cubic_2d_1o(state: np.ndarray, _: float,
                  params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a cubic polynomial for 2d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s02: Final[float] = s0 * s0
    s12: Final[float] = s1 * s1
    out[0] = (s0 * params[0]) + (s1 * params[1]) \
        + (s02 * params[2]) + (s0 * s1 * params[3]) + (s12 * params[4]) \
        + (s02 * s0 * params[5]) + (s02 * s1 * params[6]) \
        + (s0 * s12 * params[7]) + (s12 * s1 * params[8])


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __cubic_3d_1o(state: np.ndarray, _: float,
                  params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a cubic polynomial for 3d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s2: Final[float] = state[2]
    s02: Final[float] = s0 * s0
    s12: Final[float] = s1 * s1
    s22: Final[float] = s2 * s2
    out[0] = (s0 * params[0]) + (s1 * params[1]) + (s2 * params[2]) \
        + (s02 * params[3]) + (s0 * s1 * params[4]) \
        + (s0 * s2 * params[5]) + (s12 * params[6]) \
        + (s1 * s2 * params[7]) + (s22 * params[8]) \
        + (s0 * s02 * params[9]) + (s02 * s1 * params[10]) \
        + (s02 * s2 * params[11]) + (s12 * s1 * params[12]) \
        + (s12 * s0 * params[13]) + (s12 * s2 * params[14]) \
        + (s22 * s2 * params[15]) + (s22 * s0 * params[16]) \
        + (s22 * s1 * params[17])


def cubic(system: System) -> Controller:
    """
    Create a cubic controller for the given equations object.

    :param system: the equations object
    :return: the cubic controller
    """
    if system.control_dims != 1:
        raise ValueError("invalid controller dimensions "
                         f"{system.control_dims} for {system!r}.")
    name: Final[str] = "cubic"
    if system.state_dims == 2:
        return Controller(name, 2, 1, 9, __cubic_2d_1o)
    if system.state_dims == 3:
        return Controller(name, 3, 1, 19, __cubic_3d_1o)
    raise ValueError("invalid state dimensions "
                     f"{system.state_dims} for {system!r}.")



"""
An objective computing the total travel length of all teams in a game plan.

This objective function takes a game plan as argument and computes the total
travel length for all teams.
A :mod:`~moptipyapps.ttp.game_plan` is basically a matrix that, for each day
(first dimension) stores against which each team (second dimension) plays.
If a team plays at home (has a home game), its opponent is stored as a
positive number.
If the team has an away game, i.e., needs to visit the opponent, then this
opponent is stored as a negative number.
Team IDs go from `1` to `n`, i.e., a value in `1..n` indicates a home game
and a value in `-n..-1` indicates an away game.
The value `0` denotes a `bye`, i.e., that no game is scheduled for a team
at the given day.

The total game plan length is computed as follows:

1. the total length = 0
2. for each team,
    1. start at the current location = home
    2. for each day,
        1. if the opponent number is negative, the next location is the
           opponent's hometown;
        2. else if the opponent number is positive, the next location is the
           own hometown;
        3. else (if the opponent number is 0): add the `bye penalty` to the
           total length and jump to the next iteration
        4. add the distance from the current to the next location to the total
           length
        5. set the current location = the next location
    3. if the current location != own hometown, add the travel distance back
       from the current location to the hometown to the total travel length.

As penalty for the `bye` situation where no game is scheduled, we use twice
the maximum distance between any two teams plus 1.
The logic is that if a `bye` (i.e., a `0`) inserted into a game plan, it
replaces one game. Since it replaces one game, it affects up to two travels,
namely from the previous location to the game location and from the game
location to the next location.
So the optimization process could sneakily try to cut longer legs of the
tournament by inserting a `bye`.
The longest imaginable travel would be between the two cities that are
farthest away from each other and back.
By making the penalty for a `bye` exactly one distance unit longer than
this longest imaginable distance, we ensure that the travel length can
never be reduced by inserting a `bye`.
Thus, having a `bye` is always more costly than any other travel it could
replace.
"""


from typing import Final

import numba  # type: ignore
import numpy as np
from moptipy.api.objective import Objective
from moptipy.utils.logger import KeyValueLogSection
from pycommons.types import type_error

from moptipyapps.ttp.game_plan import GamePlan
from moptipyapps.ttp.instance import Instance


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def game_plan_length(
        y: np.ndarray, distances: np.ndarray, bye_penalty: int) -> int:
    """
    Compute the total travel length of a game plan.

    :param y: the game plan
    :param distances: the distance matrix
    :param bye_penalty: the penalty for `bye = 0` entries, i.e., days where
        no game is scheduled
    :returns: the total plan length

    >>> yy = np.array([[ 2, -1,  4, -3],
    ...                [-2,  1, -4,  3],
    ...                [ 3,  4, -1, -2],
    ...                [-3, -4,  1,  2],
    ...                [ 4,  3, -2, -1],
    ...                [-4, -3,  2,  1]], int)
    >>> dd = np.array([[ 0,  1,  2,  3],
    ...                [ 7,  0,  4,  5],
    ...                [ 8, 10,  0,  6],
    ...                [ 9, 11, 12,  0]], int)
    >>> 0 + 1 + 7 + 2 + 8 + 3 + 9  # team 1
    30
    >>> 7 + 1 + 0 + 5 + 11 + 4 + 10  # team 2
    38
    >>> 0 + 6 + 9 + 2 + 10 + 4  # team 3
    31
    >>> 12 + 6 + 11 + 5 + 9 + 3  # team 4
    46
    >>> 30 + 38 + 31 + 46  # total sum
    145
    >>> game_plan_length(yy, dd, 0)
    145

    >>> yy[1, 0] = 0  # add a bye
    >>> 0 + 25 + 0 + 2 + 8 + 3 + 9  # team 1
    47
    >>> game_plan_length(yy, dd, 2 * 12 + 1)
    162
    """
    days, teams = y.shape  # get the number of days and teams
    length: int = 0

    for team in range(teams):  # for each team
        current_location: int = team  # start at home
        for day in range(days):  # for each day
            next_location: int = y[day, team]
            if next_location < 0:  # away game at other team
                next_location = (-next_location) - 1
            elif next_location > 0:  # home game at home
                next_location = team
            else:  # bye = no game = penalty
                length += bye_penalty
                continue  # team stays in current location
            if current_location == next_location:
                continue  # no move
            length += distances[current_location, next_location]
            current_location = next_location

        if current_location != team:  # go back home
            length += distances[current_location, team]

    return int(length)


class GamePlanLength(Objective):
    """
    Compute the total travel length of a game plan.

    This objective function sums up all the travel lengths over all teams.
    Days without game (`bye`) are penalized.
    """

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the game plan length objective function.

        :param instance: the TTP instance
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        super().__init__()

        #: the TTP instance
        self.instance: Final[Instance] = instance
        #: the bye penalty
        self.bye_penalty: Final[int] = (2 * int(instance.max())) + 1

    def evaluate(self, x: GamePlan) -> int:
        """
        Count the errors in a game plan as objective value.

        :param x: the game plan
        :return: the number of errors in the plan
        """
        return game_plan_length(x, x.instance, self.bye_penalty)

    def lower_bound(self) -> int:
        """
        Obtain the lower bound for the travel length.

        :return: `0`
        """
        return 0

    def upper_bound(self) -> int:
        """
        Compute upper bound for the travel length: All `n*days*bye_penalty`.

        :returns: `n * days * self.bye_penalty`
        """
        n: Final[int] = self.instance.n_cities
        rounds: Final[int] = self.instance.rounds
        days: Final[int] = (n - 1) * rounds
        return n * days * self.bye_penalty

    def is_always_integer(self) -> bool:
        """
        State that this objective function is always integer-valued.

        :return: `True`
        """
        return True

    def __str__(self) -> str:
        """Get the name of this objective function."""
        return "planLength"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the instance to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("byePenalty", self.bye_penalty)



"""
A (1+1) EA for the TSP using the reversal move.

A (1+1) EA is basically a local search :mod:`~moptipy.algorithms.so.rls`
algorithm that starts with a random solution as the current solution `x`. In
each iteration, it samples one new solution from the neighborhood of `x` and
accepts it (as the new `x`), if it is better or equally good.
The algorithm here applies the reversal move, i.e., a two-opt move, as the
unary search operator used for sampling the neighborhood of `x`. The
interesting thing about this operator, when applied to the TSP, is that we
can compute the tour length of the new solution in `O(1)` from the tour length
of the current solution. This means we can very quickly decide whether the
move would be acceptable - and only apply it (in `O(n)`) if its.

The algorithm implemented here is the same as the basic (1+1) EA with `rev`
operator in the paper [1].

The original version of this code has been contributed by Mr. Tianyu LIANG
(梁天宇), <liangty@stu.hfuu.edu.cn> a Master's student at the Institute of
Applied Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School
of Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei
University (合肥学院) in Hefei, Anhui, China (中国安徽省合肥市) under the
supervision of Prof. Dr. Thomas Weise (汤卫思教授).

1. Tianyu Liang, Zhize Wu, Jörg Lässig, Daan van den Berg, and Thomas Weise.
   Solving the Traveling Salesperson Problem using Frequency Fitness
   Assignment. In Hisao Ishibuchi, Chee-Keong Kwoh, Ah-Hwee Tan, Dipti
   Srinivasan, Chunyan Miao, Anupam Trivedi, and Keeley A. Crockett, editors,
   *Proceedings of the IEEE Symposium on Foundations of Computational
   Intelligence (IEEE FOCI'22)*, part of the *IEEE Symposium Series on
   Computational Intelligence (SSCI 2022)*. December 4-7, 2022, Singapore,
   pages 360-367. IEEE. https://doi.org/10.1109/SSCI51031.2022.10022296.
2. Pedro Larrañaga, Cindy M. H. Kuijpers, Roberto H. Murga, Iñaki Inza, and
   S. Dizdarevic. Genetic Algorithms for the Travelling Salesman Problem: A
   Review of Representations and Operators. *Artificial Intelligence Review,*
   13(2):129-170, April 1999. Kluwer Academic Publishers, The Netherlands.
   https://doi.org/10.1023/A:1006529012972.
"""

from typing import Callable, Final, cast

import numba  # type: ignore
import numpy as np
from moptipy.api.algorithm import Algorithm
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection
from numpy.random import Generator
from pycommons.types import type_error

from moptipyapps.shared import SCOPE_INSTANCE
from moptipyapps.tsp.instance import Instance


@numba.njit(nogil=True, cache=True, inline="always", fastmath=False,
            boundscheck=False)
def rev_if_not_worse(i: int, j: int, n_cities: int, dist: np.ndarray,
                     x: np.ndarray, y: int) -> int:
    """
    Check a reversal move, apply it if it is better, return new tour length.

    :param i: the first, smaller index
    :param j: the second, larger index
    :param n_cities: the number of cities
    :param dist: the problem instance
    :param x: the candidate solution
    :param y: the tour length
    """
    xi: Final[int] = x[i]  # the value of x at index i
    xim1: Final[int] = x[i - 1]  # the value of x at index i-1, wraps at 0
    xj: Final[int] = x[j]  # the value of x at index j
    xjp1: Final[int] = x[(j + 1) % n_cities]  # x[j + 1] with index wrap

    # compute the difference in tour length if we would apply the move
    dy: Final[int] = (dist[xim1, xj] + dist[xi, xjp1]
                      - dist[xim1, xi] - dist[xj, xjp1])
    if dy <= 0:  # this is not a worsening improving move? ... so apply it
        # reverse the sequence from i to j in the solution
        if i == 0:  # deal with the special case that i==0
            x[0:j + 1:1] = x[j::-1]
        else:  # the normal case that i > 0
            x[i:j + 1:1] = x[j:i - 1:-1]
        return int(y + dy)  # return new tour length
    return y  # return old tour length


class TSPEA1p1revn(Algorithm):
    """A (1+1) EA using the reversal operator for the TSP."""

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the RLS algorithm for the TSP with reversal move.

        :param instance: the problem instance
        """
        super().__init__()
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        self.instance: Final[Instance] = instance

    def solve(self, process: Process) -> None:
        """
        Apply an RLS = (1+1) EA optimization process with reversing operator.

        :param process: the process instance which provides random numbers,
            functions for creating, copying, and evaluating solutions, as well
            as the termination criterion
        """
        random: Final[Generator] = process.get_random()
        # set up the fast calls
        register: Final[Callable[[np.ndarray, int], int]] =\
            cast(Callable[[np.ndarray, int], int], process.register)
        should_terminate: Final[Callable[[], bool]] = process.should_terminate
        ri: Final[Callable[[int], int]] = random.integers

        instance: Final[Instance] = self.instance  # get the instance
        n: Final[int] = instance.n_cities  # get the number of cities
        x: Final[np.ndarray] = process.create()  # create the solution
        x[:] = range(n)  # fill array with 0..n
        random.shuffle(x)  # randomly generate an initial solution

        y: int = cast(int, process.evaluate(x))  # get length of first tour
        nm1: Final[int] = n - 1  # need n-1 in the loop for the random numbers
        nm2: Final[int] = n - 2  # we need this to check the move indices
        while not should_terminate():
            i = ri(nm1)  # get the first index
            j = ri(nm1)  # get the second index
            if i > j:  # ensure that i <= j
                i, j = j, i  # swap indices if i > j
            if (i == j) or ((i == 0) and (j == nm2)):
                continue  # either a nop or a complete reversal
            y = rev_if_not_worse(i, j, n, instance, x, y)  # apply the move
            register(x, y)  # register the objective value

    def __str__(self):
        """
        Get the name of this algorithm.

        This name is then used in the directory path and file name of the
        log files.

        :returns: "tsp_ea1p1_revn"
        :retval "tsp_ea1p1_revn": always
        """
        return "tsp_ea1p1_revn"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_INSTANCE) as kv:
            self.instance.log_parameters_to(kv)



"""
The Quadratic Assignment Problem (QAP).

The quadratic assignment problem represents assignments of facilities to
locations. Between each pair of facilities, there is a flow of goods. Between
each two locations, there is a distance. The goal is to assign facilities to
locations such that the overall sum of the products of distance and flow gets
minimized. Each instance therefore presents a matrix with
:attr:`~moptipyapps.qap.instance.Instance.distances` and a matrix with flows
:attr:`~moptipyapps.qap.instance.Instance.flows`. The
:mod:`~moptipyapps.qap.objective` is then to minimize said product sum.

1. Eliane Maria Loiola, Nair Maria Maia de Abreu, Paulo Oswaldo
   Boaventura-Netto, Peter Hahn, and Tania Querido. A survey for the
   Quadratic Assignment Problem. European Journal of Operational Research.
   176(2):657-690. January 2007. https://doi.org/10.1016/j.ejor.2005.09.032.
2. Rainer E. Burkard, Eranda Çela, Panos M. Pardalos, and
   Leonidas S. Pitsoulis. The Quadratic Assignment Problem. In Ding-Zhu Du,
   Panos M. Pardalos, eds., Handbook of Combinatorial Optimization,
   pages 1713-1809, 1998, Springer New York, NY, USA.
   https://doi.org/10.1007/978-1-4613-0303-9_27.

This is code is part of the research work of Mr. Jiayang Chen (陈嘉阳), a
Master's student at the Institute of Applied Optimization (应用优化研究所,
http://iao.hfuu.edu.cn) of the School of Artificial Intelligence and Big Data
(人工智能与大数据学院) at Hefei University (合肥学院) in Hefei, Anhui, China
(中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).
"""



"""
An example experiment for dynamic control using surrogate system models.

In this experiment, we again try to synthesize (i.e., parameterize)
controllers (:mod:`~moptipyapps.dynamic_control.controller`) that steer a
dynamic system (:mod:`~moptipyapps.dynamic_control.system`) into a state by
using a figure of merit (:mod:`~moptipyapps.dynamic_control.objective`) which
minimizes both the squared system state and controller effort.

The difference compared to :mod:`~moptipyapps.dynamic_control.experiment_raw`
is that we also try to synthesize a system model at the same time. We employ
the procedure detailed in
:mod:`~moptipyapps.dynamic_control.surrogate_optimizer` for this purpose.

Word of advice: This experiment takes **extremely long** and needs
**a lot of memory**!

The starting points of the work here were conversations with Prof. Dr. Bernd
NOACK and Guy Yoslan CORNEJO MACEDA of the Harbin Institute of Technology in
Shenzhen, China (哈尔滨工业大学(深圳)).
"""

import argparse
from os.path import basename, dirname
from typing import Any, Callable, Final, Iterable, cast

import numpy as np
from moptipy.api.execution import Execution
from moptipy.api.experiment import Parallelism, run_experiment
from moptipy.api.process import Process
from moptipy.spaces.vectorspace import VectorSpace
from pycommons.io.path import Path, directory_path

from moptipyapps.dynamic_control.controllers.ann import make_ann
from moptipyapps.dynamic_control.instance import Instance
from moptipyapps.dynamic_control.objective import (
    FigureOfMerit,
    FigureOfMeritLE,
)
from moptipyapps.dynamic_control.surrogate_optimizer import (
    SurrogateOptimizer,
    _bpcmaes,
)
from moptipyapps.dynamic_control.system_model import SystemModel
from moptipyapps.dynamic_control.systems.lorenz import LORENZ_4
from moptipyapps.dynamic_control.systems.stuart_landau import STUART_LANDAU_4
from moptipyapps.dynamic_control.systems.three_coupled_oscillators import (
    THREE_COUPLED_OSCILLATORS,
)
from moptipyapps.shared import moptipyapps_argparser


def make_instances() -> Iterable[Callable[[], SystemModel]]:
    """
    Create the instances to be used in the dynamic control experiment.

    :return: the instances to be used in the dynamic control experiment.
    """
    res: list[Callable[[], SystemModel]] = []
    for system in [STUART_LANDAU_4, LORENZ_4, THREE_COUPLED_OSCILLATORS]:
        sd: int = system.state_dims
        sdp2: int = sd + 2
        ctrl_dims: int = system.control_dims
        controllers = [
            make_ann(sd, ctrl_dims, [sd, sd]),
            make_ann(sd, ctrl_dims, [sdp2, sdp2])]
        for controller in controllers:
            for ann_model in [[sd, sd, sd], [sd, sd, sd, sd],
                              [sdp2, sdp2, sdp2]]:
                res.append(cast(
                    Callable[[], SystemModel],
                    lambda _s=system, _c=controller, _m=make_ann(
                        sd + system.control_dims, sd, ann_model):
                    SystemModel(_s, _c, _m)))
    return res


#: the total objective function evaluations
MAX_FES: Final[int] = 64


def base_setup(instance: Instance) -> tuple[
        Execution, FigureOfMerit, VectorSpace]:
    """
    Create the basic setup.

    :param instance: the instance to use
    :return: the basic execution
    """
    objective: Final[FigureOfMerit] = FigureOfMeritLE(
        instance, isinstance(instance, SystemModel))
    space = instance.controller.parameter_space()
    return Execution().set_max_fes(MAX_FES).set_log_all_fes(True)\
        .set_objective(objective).set_solution_space(space), objective, space


def cmaes_raw(instance: Instance) -> Execution:
    """
    Create the Bi-Pop-CMA-ES setup.

    :param instance: the problem instance
    :return: the setup
    """
    execution, _, space = base_setup(instance)
    return execution.set_algorithm(_bpcmaes(space))


def cmaes_surrogate(instance: SystemModel,
                    fes_for_warmup: int = 16,
                    fes_for_training: int = 128,
                    fes_per_model_run: int = 128,
                    fancy_logs: bool = True) -> Execution:
    """
    Create the Bi-Pop-CMA-ES setup.

    :param instance: the problem instance
    :param fes_for_warmup: the FEs to be used for warmup
    :param fes_for_training: the milliseconds for training
    :param fes_per_model_run: the milliseconds per model run
    :param fancy_logs: should we do fancy logging?
    :return: the setup
    """
    execution, objective, space = base_setup(instance)

    return execution.set_solution_space(space).set_algorithm(
        SurrogateOptimizer(
            system_model=instance,
            controller_space=space,
            objective=objective,
            fes_for_warmup=fes_for_warmup,
            fes_for_training=fes_for_training,
            fes_per_model_run=fes_per_model_run,
            fancy_logs=fancy_logs,
            warmup_algorithm=_bpcmaes,
            model_training_algorithm=_bpcmaes,
            controller_training_algorithm=_bpcmaes))


def on_completion(instance: Any, log_file: Path, process: Process) -> None:
    """
    Plot the corresponding figures and print the objective value components.

    :param instance: the problem instance
    :param log_file: the log file
    :param process: the process
    """
    inst: Final[SystemModel] = cast(SystemModel, instance)
    dest_dir: Final[Path] = directory_path(dirname(log_file))
    base_name: str = basename(log_file)
    base_name = base_name[:base_name.rindex(".")]
    result: np.ndarray = cast(np.ndarray, process.create())
    process.get_copy_of_best_x(result)
    j: Final[float] = process.get_best_f()
    inst.describe_parameterization(f"F = {j}", result, base_name, dest_dir)


def run(base_dir: str, n_runs: int = 64) -> None:
    """
    Run the experiment.

    :param base_dir: the base directory
    :param n_runs: the number of runs
    """
    use_dir: Final[Path] = Path(base_dir)
    use_dir.ensure_dir_exists()
    instances: Final[Iterable[Callable[[], SystemModel]]] = make_instances()
    keep_instances: Final[list[Callable[[], Instance]]] = []
    raw_names: Final[set[str]] = set()
    for maker in instances:
        inst: Instance = maker()
        raw: Instance = Instance(inst.system, inst.controller)
        name: str = str(raw)
        if name in raw_names:
            continue
        raw_names.add(name)
        raw.system.describe_system_without_control(use_dir, True)
        raw.system.plot_points(use_dir, True)
        keep_instances.append(cast(Callable[[], Instance], lambda _i=raw: _i))

    run_experiment(
        base_dir=use_dir,
        instances=keep_instances,
        setups=[cmaes_raw],
        n_runs=n_runs,
        n_threads=Parallelism.PERFORMANCE,
        perform_warmup=False,
        perform_pre_warmup=False,
        on_completion=on_completion)

    setups: list[Callable[[Any], Execution]] = []
    total_training_fes: Final[int] = (MAX_FES - 1) * (2 ** 15)
    total_on_model_fes: Final[int] = (MAX_FES - 1) * 2048

    fe_choices: set[int] = {2 ** wfb for wfb in range(6)}
    fe_choices.update(MAX_FES - (2 ** wfb) for wfb in range(6))

    for warmup_fes in fe_choices:
        if not (0 < warmup_fes < MAX_FES):
            continue
        training_fes: int = max(1, int(0.5 + (total_training_fes / (
            MAX_FES - warmup_fes))))
        on_model_fes: int = max(1, int(0.5 + (total_on_model_fes / (
            MAX_FES - warmup_fes))))

        setups.append(cast(
            Callable[[Any], Execution],
            lambda i, __w=warmup_fes, __t=training_fes, __o=on_model_fes:
            cmaes_surrogate(i, __w, __t, __o, True)))

    for runs in sorted({3, 5, 7, 11, 17, 23, 31, 51, n_runs}):
        if runs > n_runs:
            break
        run_experiment(
            base_dir=use_dir,
            instances=instances,
            setups=setups,
            n_runs=runs,
            n_threads=Parallelism.ACCURATE_TIME_MEASUREMENTS,
            perform_warmup=False,
            perform_pre_warmup=False,
            on_completion=on_completion)


# Run the experiment from the command line
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = moptipyapps_argparser(
        __file__, "Dynamic Control",
        "Run the dynamic control surrogate model experiment.")
    parser.add_argument(
        "dest", help="the directory to store the experimental results under",
        type=Path, nargs="?", default="./results/")
    args: Final[argparse.Namespace] = parser.parse_args()
    run(args.dest)



"""A simple code generator."""

from io import StringIO
from typing import Any, Callable, Final

import numba  # type: ignore
import numpy as np
from pycommons.types import type_error


class CodeGenerator:
    """A simple code generator."""

    def __init__(self, args: str = "", retval: str = "None",
                 fastmath: bool = True) -> None:
        """
        Initialize the code generator.

        :param args: the command line arguments
        :param retval: the return type
        :param fastmath: do we use numba fast math?
        """
        if not isinstance(args, str):
            raise type_error(args, "args", str)
        if not isinstance(retval, str):
            raise type_error(retval, "retval", str)
        if not isinstance(fastmath, bool):
            raise type_error(fastmath, "fastmath", bool)

        io: Final[StringIO] = StringIO()
        wrt: Final[Callable[[str], int]] = io.write

        #: the callable for writing
        self.__write: Final[Callable[[str], int]] = wrt
        wrt("@numba.njit(cache=False, inline='always', fastmath"
            f"={fastmath}, boundscheck=False)\n")
        wrt(f"def ____func({args}) -> {retval}:\n")

        #: the result callable
        self.__res: Final[Callable[[], str]] = io.getvalue
        #: the current indent
        self.__indent: int = 1
        #: are we at the start of a line?
        self.__start: bool = True

    def write(self, text: str) -> None:
        """
        Write some code.

        :param text: the code text
        """
        if self.__start:
            self.__write(self.__indent * "    ")
            self.__start = False
        self.__write(text)

    def indent(self) -> None:
        """Increase the indent."""
        self.__indent += 1

    def unindent(self) -> None:
        """Increase the indent."""
        self.__indent -= 1
        if self.__indent <= 0:
            raise ValueError("indent becomes 0, not allowed")

    def writeln(self, text: str = "") -> None:
        """
        End a line.

        :param text: the text to be written
        """
        self.write(text)
        self.endline()

    def endline(self) -> None:
        """End a line."""
        if not self.__start:
            self.__write("\n")
            self.__start = True

    def build(self) -> Callable:
        """
        Compile the generated code.

        :returns: the generated function
        """
        loca: Final[dict[str, Any]] = {}
        self.endline()
        code: Final[str] = self.__res()
        globs: Final[dict[str, Any]] = {"numba": numba, "np": np,
                                        "Final": Final}
        # pylint: disable=W0122
        exec(code, globs, loca)  # nosec # nosemgrep # noqa
        res = loca["____func"]
        if not callable(res):
            raise type_error(res, f"compiled {code!r}", call=True)
        return res



"""In this file, we provide a set of best-known solutions for the 2D-BPP."""

from dataclasses import dataclass
from importlib import resources  # nosem
from math import inf, isfinite
from re import Pattern
from re import compile as p_compile
from statistics import mean
from typing import Any, Callable, Final, Iterable, Mapping, cast

from pycommons.ds.immutable_map import immutable_mapping
from pycommons.io.path import UTF8
from pycommons.math.int_math import try_int, try_int_mul
from pycommons.strings.chars import WHITESPACE
from pycommons.strings.tools import replace_regex
from pycommons.types import check_int_range

from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.packing_result import (
    BIN_COUNT_NAME,
    PackingResult,
)


def __load_references() -> Mapping[str, str]:
    """
    Load the references from the BibTeX resource.

    This is not a fully-fledged BibTeX parser. It is a very simple and crude
    method to extract bibliography keys and data.

    :return: the immutable reference dictionary, with keys sorted
        alphabetically
    """
    current_lines: list[str] = []
    key: str | None = None
    mode: int = 0
    found: Final[dict[str, str]] = {}
    single_ws: Final[Pattern] = p_compile(f"[{WHITESPACE}]")
    multi_ws: Final[Pattern] = p_compile(f"[{WHITESPACE}][{WHITESPACE}]+")
    with resources.files(__package__).joinpath(
            "bks.bib").open("r", encoding=UTF8) as stream:
        for rline in stream:
            line: str = str.rstrip(rline)
            if str.__len__(line) <= 0:
                continue
            if mode == 0:
                line = replace_regex(single_ws, "", line)
                if str.startswith(line, "@"):
                    mode = 1
                    start_idx: int = line.index("{")
                    end_idx: int = line.index(",", start_idx + 1)
                    key = str.strip(line[start_idx + 1:end_idx])
                    if str.__len__(key) <= 0:
                        raise ValueError(f"Invalid key {key!r} in {line!r}.")
                    current_lines.append(line)
                    if key in found:
                        raise ValueError(f"Duplicate key {key!r}.")
            elif mode == 1:
                if str.strip(line) in ("}", "},"):
                    mode = 0
                    current_lines.append("}")
                    found[key] = str.strip("\n".join(current_lines))
                    current_lines.clear()
                else:
                    current_lines.append(replace_regex(multi_ws, " ", line))
    if dict.__len__(found) <= 0:
        raise ValueError("Found no references!")
    return immutable_mapping({k: found[k] for k in sorted(found.keys())})


#: the references to literature
_REFERENCES: Final[Mapping[str, str]] = __load_references()

#: a constant for no reference
NO_REF: Final[str] = "{{NO_REF}}"


@dataclass(frozen=True, init=False, order=True, eq=True)
class Element:
    """The algorithm or instance group specification."""

    #: the name or name prefix
    name: str
    #: the name suffix, if any, and the empty string otherwise
    name_suffix: str
    #: the reference to the related work
    reference: str

    def __init__(self, name: str,
                 name_suffix: str, reference: str = NO_REF) -> None:
        """
        Initialize this element.

        :param name: the name or name prefix
        :param name_suffix: the name suffix, or an empty string
        :param reference: the reference
        """
        name = str.strip(name)
        if str.__len__(name) <= 0:
            raise ValueError(f"Invalid name {name!r}.")
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "name_suffix", str.strip(name_suffix))
        reference = str.strip(reference)
        for ref in reference.split(","):
            if ref == NO_REF:
                continue
            if ref not in _REFERENCES:
                raise ValueError(f"Invalid reference {ref!r}.")
        object.__setattr__(self, "reference", reference)

    def get_bibtex(self) -> Iterable[str]:
        """
        Get the BibTeX for this element.

        :return: the BibTeX string
        """
        return (_REFERENCES[r] for r in self.reference.split(","))


#: the BRKGA for 2D bin packing with rotation
BRKGA_BPP_2R: Final[Element] = Element(
    "BRKGA", "BRKGABPPRTR", "GR2013ABRKGAF2A3BPP")
#: the BRKGA for 2D bin packing without rotation
BRKGA_BPP_ANB: Final[Element] = Element(
    "BRKGA", "BRKGABPPRANB", "GR2013ABRKGAF2A3BPP")
#: the grasp/vnd for 2D bin packing without rotation
GRASP_VND: Final[Element] = Element(
    "GRASPVNDGRASP", "GRASPVNDVND", "PAVOT2010AHGVAFTATDBP")
#: the price-and-cut algorithm for 2D bin packing with rotation
PAC: Final[Element] = Element("PAC", "", "CGRS2020PACATSMTOOSFT2BPP")
#: the HHANO-R algorithm for 2D bin packing with rotation
HHANO_R: Final[Element] = Element("HHANO", "HHANOR", "BDC2015RHHAFTOONO2BPP")
#: the HHANO-SR algorithm for 2D bin packing with rotation
HHANO_SR: Final[Element] = Element("HHANO", "HHANOSR", "BDC2015RHHAFTOONO2BPP")
#: the MXGA algorithm for 2D bin packing with rotation
MXGA: Final[Element] = Element("MXGA", "", "L2008AGAFTDBPP")
#: the LGFi algorithm
EALGFI: Final[Element] = Element(
    "EALGFIEA", "EALGFILGFI", "BS2013ST2BPPBMOAHEA")

#: the small A instances
A_SMALL: Final[Element] = Element("a", "small", "MAVdC2010AFMFTTDGCSP")
#: the median sized A instances
A_MED: Final[Element] = Element("a", "med", "MAVdC2010AFMFTTDGCSP")
#: the large A instances
A_LARGE: Final[Element] = Element("a", "large", "MAVdC2010AFMFTTDGCSP")
#: the asqas
ASQAS: Final[Element] = Element("asqas", "", "vdBBMSB2016ASIASSTFI")
#: the first part of the Beng instances
BENG_1_8: Final[Element] = Element("beng", "1-8", "B1982PRPAHA")
#: the second part of the Beng instances
BENG_9_10: Final[Element] = Element("beng", "9-10", "B1982PRPAHA")
#: the class 1-20 benchmarks
CLASS_1_20: Final[Element] = Element(
    "class 1", "20", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 1-40 benchmarks
CLASS_1_40: Final[Element] = Element(
    "class 1", "40", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 1-60 benchmarks
CLASS_1_60: Final[Element] = Element(
    "class 1", "60", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 1-80 benchmarks
CLASS_1_80: Final[Element] = Element(
    "class 1", "80", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 1-100 benchmarks
CLASS_1_100: Final[Element] = Element(
    "class 1", "100", "BW1987TDFBPA,MV1998ESOTTDFBPP")

#: the class 2-20 benchmarks
CLASS_2_20: Final[Element] = Element(
    "class 2", "20", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 2-40 benchmarks
CLASS_2_40: Final[Element] = Element(
    "class 2", "40", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 2-60 benchmarks
CLASS_2_60: Final[Element] = Element(
    "class 2", "60", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 2-80 benchmarks
CLASS_2_80: Final[Element] = Element(
    "class 2", "80", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 1-100 benchmarks
CLASS_2_100: Final[Element] = Element(
    "class 2", "100", "BW1987TDFBPA,MV1998ESOTTDFBPP")

#: the class 3-20 benchmarks
CLASS_3_20: Final[Element] = Element(
    "class 3", "20", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 1-40 benchmarks
CLASS_3_40: Final[Element] = Element(
    "class 3", "40", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 3-60 benchmarks
CLASS_3_60: Final[Element] = Element(
    "class 3", "60", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 3-80 benchmarks
CLASS_3_80: Final[Element] = Element(
    "class 3", "80", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 3-100 benchmarks
CLASS_3_100: Final[Element] = Element(
    "class 3", "100", "BW1987TDFBPA,MV1998ESOTTDFBPP")

#: the class 4-20 benchmarks
CLASS_4_20: Final[Element] = Element(
    "class 4", "20", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 4-40 benchmarks
CLASS_4_40: Final[Element] = Element(
    "class 4", "40", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 4-60 benchmarks
CLASS_4_60: Final[Element] = Element(
    "class 4", "60", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 4-80 benchmarks
CLASS_4_80: Final[Element] = Element(
    "class 4", "80", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 4-100 benchmarks
CLASS_4_100: Final[Element] = Element(
    "class 4", "100", "BW1987TDFBPA,MV1998ESOTTDFBPP")

#: the class 5-20 benchmarks
CLASS_5_20: Final[Element] = Element(
    "class 5", "20", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 5-40 benchmarks
CLASS_5_40: Final[Element] = Element(
    "class 5", "40", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 5-60 benchmarks
CLASS_5_60: Final[Element] = Element(
    "class 5", "60", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 5-80 benchmarks
CLASS_5_80: Final[Element] = Element(
    "class 5", "80", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 5-100 benchmarks
CLASS_5_100: Final[Element] = Element(
    "class 5", "100", "BW1987TDFBPA,MV1998ESOTTDFBPP")

#: the class 6-20 benchmarks
CLASS_6_20: Final[Element] = Element(
    "class 6", "20", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 6-40 benchmarks
CLASS_6_40: Final[Element] = Element(
    "class 6", "40", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 6-60 benchmarks
CLASS_6_60: Final[Element] = Element(
    "class 6", "60", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 6-80 benchmarks
CLASS_6_80: Final[Element] = Element(
    "class 6", "80", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 6-100 benchmarks
CLASS_6_100: Final[Element] = Element(
    "class 6", "100", "BW1987TDFBPA,MV1998ESOTTDFBPP")

#: the class 7-20 benchmarks
CLASS_7_20: Final[Element] = Element(
    "class 7", "20", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 7-40 benchmarks
CLASS_7_40: Final[Element] = Element(
    "class 7", "40", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 7-60 benchmarks
CLASS_7_60: Final[Element] = Element(
    "class 7", "60", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 7-80 benchmarks
CLASS_7_80: Final[Element] = Element(
    "class 7", "80", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 7-100 benchmarks
CLASS_7_100: Final[Element] = Element(
    "class 7", "100", "BW1987TDFBPA,MV1998ESOTTDFBPP")

#: the class 8-20 benchmarks
CLASS_8_20: Final[Element] = Element(
    "class 8", "20", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 8-40 benchmarks
CLASS_8_40: Final[Element] = Element(
    "class 8", "40", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 8-60 benchmarks
CLASS_8_60: Final[Element] = Element(
    "class 8", "60", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 8-80 benchmarks
CLASS_8_80: Final[Element] = Element(
    "class 8", "80", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 8-100 benchmarks
CLASS_8_100: Final[Element] = Element(
    "class 8", "100", "BW1987TDFBPA,MV1998ESOTTDFBPP")

#: the class 9-20 benchmarks
CLASS_9_20: Final[Element] = Element(
    "class 9", "20", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 9-40 benchmarks
CLASS_9_40: Final[Element] = Element(
    "class 9", "40", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 9-60 benchmarks
CLASS_9_60: Final[Element] = Element(
    "class 9", "60", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 9-80 benchmarks
CLASS_9_80: Final[Element] = Element(
    "class 9", "80", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 9-100 benchmarks
CLASS_9_100: Final[Element] = Element(
    "class 9", "100", "BW1987TDFBPA,MV1998ESOTTDFBPP")

#: the class 10-20 benchmarks
CLASS_10_20: Final[Element] = Element(
    "class 10", "20", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 10-40 benchmarks
CLASS_10_40: Final[Element] = Element(
    "class 10", "40", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 10-60 benchmarks
CLASS_10_60: Final[Element] = Element(
    "class 10", "60", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 10-80 benchmarks
CLASS_10_80: Final[Element] = Element(
    "class 10", "80", "BW1987TDFBPA,MV1998ESOTTDFBPP")
#: the class 6-100 benchmarks
CLASS_10_100: Final[Element] = Element(
    "class 10", "100", "BW1987TDFBPA,MV1998ESOTTDFBPP")


def __sort_key(txt: str) -> list[str]:
    """
    Get a sort key from a string.

    :param txt: the string
    :return: the sort key
    """
    kys: list[str] = []
    for v in txt.split(","):
        for vv in v.split("-"):
            for vvv in vv.split("_"):
                for vvvv in vvv.split(" "):
                    if vvvv == "small":
                        kys.append("1_small")
                    elif vvvv == "med":
                        kys.append("2_med")
                    elif vvvv == "large":
                        kys.append("3_large")
                    elif vvvv.startswith("asqas"):
                        kys.append(f"zzzzzzzzzzzzzzzzzz{vvvv}")
                    else:
                        kys.append(vvvv.rjust(10, "0")
                                   if vvvv.isdigit() else vvvv)
    return kys


#: The set of class and beng instance groups
CLASS_AND_BENG: Final[tuple[Element, ...]] = (
    BENG_1_8, BENG_9_10, CLASS_1_20, CLASS_1_40, CLASS_1_60, CLASS_1_80,
    CLASS_1_100, CLASS_2_20, CLASS_2_40, CLASS_2_60, CLASS_2_80,
    CLASS_2_100, CLASS_3_20, CLASS_3_40, CLASS_3_60, CLASS_3_80,
    CLASS_3_100, CLASS_4_20, CLASS_4_40, CLASS_4_60, CLASS_4_80,
    CLASS_4_100, CLASS_5_20, CLASS_5_40, CLASS_5_60, CLASS_5_80,
    CLASS_5_100, CLASS_6_20, CLASS_6_40, CLASS_6_60, CLASS_6_80,
    CLASS_6_100, CLASS_7_20, CLASS_7_40, CLASS_7_60, CLASS_7_80,
    CLASS_7_100, CLASS_8_20, CLASS_8_40, CLASS_8_60, CLASS_8_80,
    CLASS_8_100, CLASS_9_20, CLASS_9_40, CLASS_9_60, CLASS_9_80,
    CLASS_9_100, CLASS_10_20, CLASS_10_40, CLASS_10_60, CLASS_10_80,
    CLASS_10_100,
)


def __make_instance_map() -> Mapping[Element, tuple[str, ...]]:
    """
    Make an instance map.

    :return: the instance map.
    """
    groups: Final[list[Element]] = [
        A_SMALL, A_MED, A_LARGE, ASQAS,
        BENG_1_8, BENG_9_10, CLASS_1_20, CLASS_1_40, CLASS_1_60, CLASS_1_80,
        CLASS_1_100, CLASS_2_20, CLASS_2_40, CLASS_2_60, CLASS_2_80,
        CLASS_2_100, CLASS_3_20, CLASS_3_40, CLASS_3_60, CLASS_3_80,
        CLASS_3_100, CLASS_4_20, CLASS_4_40, CLASS_4_60, CLASS_4_80,
        CLASS_4_100, CLASS_5_20, CLASS_5_40, CLASS_5_60, CLASS_5_80,
        CLASS_5_100, CLASS_6_20, CLASS_6_40, CLASS_6_60, CLASS_6_80,
        CLASS_6_100, CLASS_7_20, CLASS_7_40, CLASS_7_60, CLASS_7_80,
        CLASS_7_100, CLASS_8_20, CLASS_8_40, CLASS_8_60, CLASS_8_80,
        CLASS_8_100, CLASS_9_20, CLASS_9_40, CLASS_9_60, CLASS_9_80,
        CLASS_9_100, CLASS_10_20, CLASS_10_40, CLASS_10_60, CLASS_10_80,
        CLASS_10_100]

    data: dict[Element, tuple[str, ...]] = {}
    for group in Instance.list_resources_groups():
        prefix: str = group[0]
        suffix: str = "" if group[1] is None else group[1]
        insts: tuple[str, ...] = group[2]
        found: bool = False
        for ggg in groups:
            if (ggg.name == prefix) and (ggg.name_suffix == suffix):
                if ggg in data:
                    raise ValueError(f"Encountered {ggg} twice?")
                data[ggg] = insts
                found = True
                break
        if not found:
            raise ValueError(f"Did not find {group!r}.")

    return immutable_mapping({k: data[k] for k in sorted(
        data.keys(), key=lambda e: __sort_key(f"{e.name} {e.name_suffix}"))})


#: a mapping of instance groups to instances
GROUPS_TO_INSTANCES: Final[Mapping[Element, tuple[
    str, ...]]] = __make_instance_map()

#: A sort key function for instance groups
INST_GROUP_SORT_KEY: Final[Callable[[Element], int]] \
    = list(GROUPS_TO_INSTANCES.keys()).index


def __make_instance_to_groups() -> Mapping[str, Element]:
    """
    Make a mapping of instances to groups.

    :return: the mapping of instances to groups.
    """
    data: list[tuple[str, Element]] = [(
        v, k) for k, vv in GROUPS_TO_INSTANCES.items() for v in vv]
    data.sort(key=lambda x: __sort_key(x[0]))
    return immutable_mapping(dict(data))


#: a mapping of instances to instance groups
INSTANCES_TO_GROUPS: Final[Mapping[str, Element]] = \
    __make_instance_to_groups()


def __lb_avg_denormalize(
        with_rotation: bool, algo: Element, group: Element,
        value: int | float, min_result: int = -1_000_000) -> tuple[
        bool, Element, Element, int]:
    """
    De-normalize using the maximum of the Dell'Amico and geometric bound.

    :param with_rotation: is this a result with or without rotation?
    :param algo: the algorithm used
    :param group: the instance group
    :param value: the value
    :param min_result: the minimum denormalized result
    :return: de-normalized average
    """
    return (with_rotation, algo, group, max(
        min_result, int(round(try_int_mul(sum(Instance.from_resource(
            ins).lower_bound_bins for ins in GROUPS_TO_INSTANCES[group]),
            value)))))


#: the related works with rotation averaged
__RW__AVERAGE: Final[tuple[tuple[
    bool, Element, Element, int], ...]] = (
    (True, BRKGA_BPP_2R, CLASS_1_20, 66),
    (False, BRKGA_BPP_ANB, CLASS_1_20, 71),
    (True, BRKGA_BPP_2R, CLASS_1_40, 128),
    (False, BRKGA_BPP_ANB, CLASS_1_40, 134),
    (True, BRKGA_BPP_2R, CLASS_1_60, 195),
    (False, BRKGA_BPP_ANB, CLASS_1_60, 200),
    (True, BRKGA_BPP_2R, CLASS_1_80, 270),
    (False, BRKGA_BPP_ANB, CLASS_1_80, 275),
    (True, BRKGA_BPP_2R, CLASS_1_100, 313),
    (False, BRKGA_BPP_ANB, CLASS_1_100, 317),
    (True, BRKGA_BPP_2R, CLASS_2_20, 10),
    (False, BRKGA_BPP_ANB, CLASS_2_20, 10),
    (True, BRKGA_BPP_2R, CLASS_2_40, 19),
    (False, BRKGA_BPP_ANB, CLASS_2_40, 19),
    (True, BRKGA_BPP_2R, CLASS_2_60, 25),
    (False, BRKGA_BPP_ANB, CLASS_2_60, 25),
    (True, BRKGA_BPP_2R, CLASS_2_80, 31),
    (False, BRKGA_BPP_ANB, CLASS_2_80, 31),
    (True, BRKGA_BPP_2R, CLASS_2_100, 39),
    (False, BRKGA_BPP_ANB, CLASS_2_100, 39),
    (True, BRKGA_BPP_2R, CLASS_3_20, 47),
    (False, BRKGA_BPP_ANB, CLASS_3_20, 51),
    (True, BRKGA_BPP_2R, CLASS_3_40, 92),
    (False, BRKGA_BPP_ANB, CLASS_3_40, 94),
    (True, BRKGA_BPP_2R, CLASS_3_60, 134),
    (False, BRKGA_BPP_ANB, CLASS_3_60, 139),
    (True, BRKGA_BPP_2R, CLASS_3_80, 182),
    (False, BRKGA_BPP_ANB, CLASS_3_80, 189),
    (True, BRKGA_BPP_2R, CLASS_3_100, 220),
    (False, BRKGA_BPP_ANB, CLASS_3_100, 223),
    (True, BRKGA_BPP_2R, CLASS_4_20, 10),
    (False, BRKGA_BPP_ANB, CLASS_4_20, 10),
    (True, BRKGA_BPP_2R, CLASS_4_40, 19),
    (False, BRKGA_BPP_ANB, CLASS_4_40, 19),
    (True, BRKGA_BPP_2R, CLASS_4_60, 23),
    (False, BRKGA_BPP_ANB, CLASS_4_60, 25),
    (True, BRKGA_BPP_2R, CLASS_4_80, 31),
    (False, BRKGA_BPP_ANB, CLASS_4_80, 31),
    (True, BRKGA_BPP_2R, CLASS_4_100, 37),
    (False, BRKGA_BPP_ANB, CLASS_4_100, 37),
    (True, BRKGA_BPP_2R, CLASS_5_20, 59),
    (False, BRKGA_BPP_ANB, CLASS_5_20, 65),
    (True, BRKGA_BPP_2R, CLASS_5_40, 114),
    (False, BRKGA_BPP_ANB, CLASS_5_40, 119),
    (True, BRKGA_BPP_2R, CLASS_5_60, 172),
    (False, BRKGA_BPP_ANB, CLASS_5_60, 180),
    (True, BRKGA_BPP_2R, CLASS_5_80, 239),
    (False, BRKGA_BPP_ANB, CLASS_5_80, 247),
    (True, BRKGA_BPP_2R, CLASS_5_100, 277),
    (False, BRKGA_BPP_ANB, CLASS_5_100, 281),
    (True, BRKGA_BPP_2R, CLASS_6_20, 10),
    (False, BRKGA_BPP_ANB, CLASS_6_20, 10),
    (True, BRKGA_BPP_2R, CLASS_6_40, 16),
    (False, BRKGA_BPP_ANB, CLASS_6_40, 16),
    (True, BRKGA_BPP_2R, CLASS_6_60, 21),
    (False, BRKGA_BPP_ANB, CLASS_6_60, 21),
    (True, BRKGA_BPP_2R, CLASS_6_80, 30),
    (False, BRKGA_BPP_ANB, CLASS_6_80, 30),
    (True, BRKGA_BPP_2R, CLASS_6_100, 32),
    (False, BRKGA_BPP_ANB, CLASS_6_100, 33),
    (True, BRKGA_BPP_2R, CLASS_7_20, 52),
    (False, BRKGA_BPP_ANB, CLASS_7_20, 55),
    (True, BRKGA_BPP_2R, CLASS_7_40, 102),
    (False, BRKGA_BPP_ANB, CLASS_7_40, 111),
    (True, BRKGA_BPP_2R, CLASS_7_60, 146),
    (False, BRKGA_BPP_ANB, CLASS_7_60, 158),
    (True, BRKGA_BPP_2R, CLASS_7_80, 208),
    (False, BRKGA_BPP_ANB, CLASS_7_80, 232),
    (True, BRKGA_BPP_2R, CLASS_7_100, 250),
    (False, BRKGA_BPP_ANB, CLASS_7_100, 271),
    (True, BRKGA_BPP_2R, CLASS_8_20, 53),
    (False, BRKGA_BPP_ANB, CLASS_8_20, 58),
    (True, BRKGA_BPP_2R, CLASS_8_40, 103),
    (False, BRKGA_BPP_ANB, CLASS_8_40, 113),
    (True, BRKGA_BPP_2R, CLASS_8_60, 147),
    (False, BRKGA_BPP_ANB, CLASS_8_60, 161),
    (True, BRKGA_BPP_2R, CLASS_8_80, 204),
    (False, BRKGA_BPP_ANB, CLASS_8_80, 224),
    (True, BRKGA_BPP_2R, CLASS_8_100, 252),
    (False, BRKGA_BPP_ANB, CLASS_8_100, 278),
    (True, BRKGA_BPP_2R, CLASS_9_20, 143),
    (False, BRKGA_BPP_ANB, CLASS_9_20, 143),
    (True, BRKGA_BPP_2R, CLASS_9_40, 275),
    (False, BRKGA_BPP_ANB, CLASS_9_40, 278),
    (True, BRKGA_BPP_2R, CLASS_9_60, 435),
    (False, BRKGA_BPP_ANB, CLASS_9_60, 437),
    (True, BRKGA_BPP_2R, CLASS_9_80, 573),
    (False, BRKGA_BPP_ANB, CLASS_9_80, 577),
    (True, BRKGA_BPP_2R, CLASS_9_100, 693),
    (False, BRKGA_BPP_ANB, CLASS_9_100, 695),
    (True, BRKGA_BPP_2R, CLASS_10_20, 41),
    (False, BRKGA_BPP_ANB, CLASS_10_20, 42),
    (True, BRKGA_BPP_2R, CLASS_10_40, 72),
    (False, BRKGA_BPP_ANB, CLASS_10_40, 74),
    (True, BRKGA_BPP_2R, CLASS_10_60, 99),
    (False, BRKGA_BPP_ANB, CLASS_10_60, 100),
    (True, BRKGA_BPP_2R, CLASS_10_80, 125),
    (False, BRKGA_BPP_ANB, CLASS_10_80, 128),
    (True, BRKGA_BPP_2R, CLASS_10_100, 154),
    (False, BRKGA_BPP_ANB, CLASS_10_100, 158),
    (True, BRKGA_BPP_2R, BENG_1_8, 54),
    (False, BRKGA_BPP_ANB, BENG_1_8, 54),
    (True, BRKGA_BPP_2R, BENG_9_10, 13),
    (False, BRKGA_BPP_ANB, BENG_9_10, 13),
    (False, PAC, CLASS_1_20, 71),
    (False, PAC, CLASS_1_40, 134),
    (False, PAC, CLASS_1_60, 200),
    (False, PAC, CLASS_1_80, 275),
    (False, PAC, CLASS_1_100, 317),
    (False, PAC, CLASS_2_20, 10),
    (False, PAC, CLASS_2_40, 19),
    (False, PAC, CLASS_2_60, 25),
    (False, PAC, CLASS_2_80, 31),
    (False, PAC, CLASS_2_100, 39),
    (False, PAC, CLASS_3_20, 51),
    (False, PAC, BENG_1_8, 54),
    (False, PAC, BENG_9_10, 13),
    (True, PAC, CLASS_1_20, 66),
    (True, PAC, CLASS_1_40, 128),
    (True, PAC, CLASS_1_60, 195),
    (True, PAC, CLASS_1_80, 270),
    (True, PAC, CLASS_1_100, 313),
    (True, PAC, CLASS_2_20, 10),
    (True, PAC, CLASS_2_40, 19),
    (True, PAC, CLASS_2_60, 25),
    (True, PAC, CLASS_2_80, 31),
    (True, PAC, CLASS_2_100, 39),
    (True, PAC, CLASS_3_20, 47),
    (True, PAC, BENG_1_8, 54),
    (True, PAC, BENG_9_10, 13),
    (True, HHANO_R, CLASS_1_20, 66),
    (True, HHANO_SR, CLASS_1_20, 66),
    (True, HHANO_R, CLASS_1_40, 131),
    (True, HHANO_SR, CLASS_1_40, 129),
    (True, HHANO_R, CLASS_1_60, 196),
    (True, HHANO_SR, CLASS_1_60, 195),
    (True, HHANO_R, CLASS_1_80, 270),
    (True, HHANO_SR, CLASS_1_80, 270),
    (True, HHANO_R, CLASS_1_100, 314),
    (True, HHANO_SR, CLASS_1_100, 313),
    (True, HHANO_R, CLASS_2_20, 10),
    (True, HHANO_SR, CLASS_2_20, 10),
    (True, HHANO_R, CLASS_2_40, 20),
    (True, HHANO_SR, CLASS_2_40, 19),
    (True, HHANO_R, CLASS_2_60, 25),
    (True, HHANO_SR, CLASS_2_60, 25),
    (True, HHANO_R, CLASS_2_80, 31),
    (True, HHANO_SR, CLASS_2_80, 31),
    (True, HHANO_R, CLASS_2_100, 39),
    (True, HHANO_SR, CLASS_2_100, 39),
    (True, HHANO_R, CLASS_3_20, 48),
    (True, HHANO_SR, CLASS_3_20, 48),
    (True, HHANO_R, CLASS_3_40, 95),
    (True, HHANO_SR, CLASS_3_40, 95),
    (True, HHANO_R, CLASS_3_60, 137),
    (True, HHANO_SR, CLASS_3_60, 137),
    (True, HHANO_R, CLASS_3_80, 186),
    (True, HHANO_SR, CLASS_3_80, 187),
    (True, HHANO_R, CLASS_3_100, 225),
    (True, HHANO_SR, CLASS_3_100, 225),
    (True, HHANO_R, CLASS_4_20, 10),
    (True, HHANO_SR, CLASS_4_20, 10),
    (True, HHANO_R, CLASS_4_40, 19),
    (True, HHANO_SR, CLASS_4_40, 19),
    (True, HHANO_R, CLASS_4_60, 25),
    (True, HHANO_SR, CLASS_4_60, 25),
    (True, HHANO_R, CLASS_4_80, 32),
    (True, HHANO_SR, CLASS_4_80, 33),
    (True, HHANO_R, CLASS_4_100, 38),
    (True, HHANO_SR, CLASS_4_100, 38),
    (True, HHANO_R, CLASS_5_20, 59),
    (True, HHANO_SR, CLASS_5_20, 59),
    (True, HHANO_R, CLASS_5_40, 116),
    (True, HHANO_SR, CLASS_5_40, 115),
    (True, HHANO_R, CLASS_5_60, 175),
    (True, HHANO_SR, CLASS_5_60, 176),
    (True, HHANO_R, CLASS_5_80, 240),
    (True, HHANO_SR, CLASS_5_80, 241),
    (True, HHANO_R, CLASS_5_100, 284),
    (True, HHANO_SR, CLASS_5_100, 284),
    (True, HHANO_R, CLASS_6_20, 10),
    (True, HHANO_SR, CLASS_6_20, 10),
    (True, HHANO_R, CLASS_6_40, 18),
    (True, HHANO_SR, CLASS_6_40, 17),
    (True, HHANO_R, CLASS_6_60, 22),
    (True, HHANO_SR, CLASS_6_60, 22),
    (True, HHANO_R, CLASS_6_80, 30),
    (True, HHANO_SR, CLASS_6_80, 30),
    (True, HHANO_R, CLASS_6_100, 34),
    (True, HHANO_SR, CLASS_6_100, 34),
    (True, HHANO_R, CLASS_7_20, 52),
    (True, HHANO_SR, CLASS_7_20, 52),
    (True, HHANO_R, CLASS_7_40, 106),
    (True, HHANO_SR, CLASS_7_40, 107),
    (True, HHANO_R, CLASS_7_60, 152),
    (True, HHANO_SR, CLASS_7_60, 153),
    (True, HHANO_R, CLASS_7_80, 216),
    (True, HHANO_SR, CLASS_7_80, 217),
    (True, HHANO_R, CLASS_7_100, 260),
    (True, HHANO_SR, CLASS_7_100, 259),
    (True, HHANO_R, CLASS_8_20, 53),
    (True, HHANO_SR, CLASS_8_20, 53),
    (True, HHANO_R, CLASS_8_40, 106),
    (True, HHANO_SR, CLASS_8_40, 105),
    (True, HHANO_R, CLASS_8_60, 155),
    (True, HHANO_SR, CLASS_8_60, 154),
    (True, HHANO_R, CLASS_8_80, 213),
    (True, HHANO_SR, CLASS_8_80, 214),
    (True, HHANO_R, CLASS_8_100, 261),
    (True, HHANO_SR, CLASS_8_100, 262),
    (True, HHANO_R, CLASS_9_20, 143),
    (True, HHANO_SR, CLASS_9_20, 143),
    (True, HHANO_R, CLASS_9_40, 275),
    (True, HHANO_SR, CLASS_9_40, 275),
    (True, HHANO_R, CLASS_9_60, 435),
    (True, HHANO_SR, CLASS_9_60, 435),
    (True, HHANO_R, CLASS_9_80, 573),
    (True, HHANO_SR, CLASS_9_80, 573),
    (True, HHANO_R, CLASS_9_100, 693),
    (True, HHANO_SR, CLASS_9_100, 693),
    (True, HHANO_R, CLASS_10_20, 41),
    (True, HHANO_SR, CLASS_10_20, 41),
    (True, HHANO_R, CLASS_10_40, 73),
    (True, HHANO_SR, CLASS_10_40, 73),
    (True, HHANO_R, CLASS_10_60, 101),
    (True, HHANO_SR, CLASS_10_60, 101),
    (True, HHANO_R, CLASS_10_80, 129),
    (True, HHANO_SR, CLASS_10_80, 130),
    (True, HHANO_R, CLASS_10_100, 161),
    (True, HHANO_SR, CLASS_10_100, 162),
    __lb_avg_denormalize(True, MXGA, CLASS_1_20, 1.03),
    __lb_avg_denormalize(True, MXGA, CLASS_1_40, 1.04),
    __lb_avg_denormalize(True, MXGA, CLASS_1_60, 1.04, 195),  # 19.4
    __lb_avg_denormalize(True, MXGA, CLASS_1_80, 1.06),
    __lb_avg_denormalize(True, MXGA, CLASS_1_100, 1.02, 313),  # 31.2
    __lb_avg_denormalize(True, MXGA, CLASS_2_20, 1.0),
    __lb_avg_denormalize(True, MXGA, CLASS_2_40, 1.1),
    __lb_avg_denormalize(True, MXGA, CLASS_2_60, 1.0),
    __lb_avg_denormalize(True, MXGA, CLASS_2_80, 1.0),
    __lb_avg_denormalize(True, MXGA, CLASS_2_100, 1.0),
    __lb_avg_denormalize(True, MXGA, CLASS_3_20, 1.04),
    __lb_avg_denormalize(True, MXGA, CLASS_3_40, 1.09),
    __lb_avg_denormalize(True, MXGA, CLASS_3_60, 1.08),
    __lb_avg_denormalize(True, MXGA, CLASS_3_80, 1.06),
    __lb_avg_denormalize(True, MXGA, CLASS_3_100, 1.06),
    __lb_avg_denormalize(True, MXGA, CLASS_4_20, 1.0),
    __lb_avg_denormalize(True, MXGA, CLASS_4_40, 1.0),
    __lb_avg_denormalize(True, MXGA, CLASS_4_60, 1.10),
    __lb_avg_denormalize(True, MXGA, CLASS_4_80, 1.07),
    __lb_avg_denormalize(True, MXGA, CLASS_4_100, 1.03),
    __lb_avg_denormalize(True, MXGA, CLASS_5_20, 1.04),
    __lb_avg_denormalize(True, MXGA, CLASS_5_40, 1.06),
    __lb_avg_denormalize(True, MXGA, CLASS_5_60, 1.07),
    __lb_avg_denormalize(True, MXGA, CLASS_5_80, 1.07),
    __lb_avg_denormalize(True, MXGA, CLASS_5_100, 1.05),
    __lb_avg_denormalize(True, MXGA, CLASS_6_20, 1.0),
    __lb_avg_denormalize(True, MXGA, CLASS_6_40, 1.4),
    __lb_avg_denormalize(True, MXGA, CLASS_6_60, 1.0),
    __lb_avg_denormalize(True, MXGA, CLASS_6_80, 1.0),
    __lb_avg_denormalize(True, MXGA, CLASS_6_100, 1.07),
    __lb_avg_denormalize(True, MXGA, CLASS_7_20, 1.11),
    __lb_avg_denormalize(True, MXGA, CLASS_7_40, 1.07),
    __lb_avg_denormalize(True, MXGA, CLASS_7_60, 1.05),
    __lb_avg_denormalize(True, MXGA, CLASS_7_80, 1.08),
    __lb_avg_denormalize(True, MXGA, CLASS_7_100, 1.07),
    __lb_avg_denormalize(True, MXGA, CLASS_8_20, 1.10),
    __lb_avg_denormalize(True, MXGA, CLASS_8_40, 1.09),
    __lb_avg_denormalize(True, MXGA, CLASS_8_60, 1.06),
    __lb_avg_denormalize(True, MXGA, CLASS_8_80, 1.07),
    __lb_avg_denormalize(True, MXGA, CLASS_8_100, 1.06),
    __lb_avg_denormalize(True, MXGA, CLASS_9_20, 1.0),
    __lb_avg_denormalize(True, MXGA, CLASS_9_40, 1.01),
    __lb_avg_denormalize(True, MXGA, CLASS_9_60, 1.01),
    __lb_avg_denormalize(True, MXGA, CLASS_9_80, 1.01),
    __lb_avg_denormalize(True, MXGA, CLASS_9_100, 1.01),
    __lb_avg_denormalize(True, MXGA, CLASS_10_20, 1.13),
    __lb_avg_denormalize(True, MXGA, CLASS_10_40, 1.06),
    __lb_avg_denormalize(True, MXGA, CLASS_10_60, 1.07),
    __lb_avg_denormalize(True, MXGA, CLASS_10_80, 1.06),
    __lb_avg_denormalize(True, MXGA, CLASS_10_100, 1.04),
    (False, EALGFI, CLASS_1_20, 71),
    (False, EALGFI, CLASS_1_40, 134),
    (False, EALGFI, CLASS_1_60, 200),
    (False, EALGFI, CLASS_1_80, 275),
    (False, EALGFI, CLASS_1_100, 317),
    (False, EALGFI, CLASS_2_20, 10),
    (False, EALGFI, CLASS_2_40, 19),
    (False, EALGFI, CLASS_2_60, 25),
    (False, EALGFI, CLASS_2_80, 31),
    (False, EALGFI, CLASS_2_100, 39),
    (False, EALGFI, CLASS_3_20, 51),
    (False, EALGFI, CLASS_3_40, 94),
    (False, EALGFI, CLASS_3_60, 139),
    (False, EALGFI, CLASS_3_80, 189),
    (False, EALGFI, CLASS_3_100, 224),
    (False, EALGFI, CLASS_4_20, 10),
    (False, EALGFI, CLASS_4_40, 19),
    (False, EALGFI, CLASS_4_60, 23),
    (False, EALGFI, CLASS_4_80, 31),
    (False, EALGFI, CLASS_4_100, 37),
    (False, EALGFI, CLASS_5_20, 65),
    (False, EALGFI, CLASS_5_40, 119),
    (False, EALGFI, CLASS_5_60, 180),
    (False, EALGFI, CLASS_5_80, 247),
    (False, EALGFI, CLASS_5_100, 284),
    (False, EALGFI, CLASS_6_20, 10),
    (False, EALGFI, CLASS_6_40, 17),
    (False, EALGFI, CLASS_6_60, 21),
    (False, EALGFI, CLASS_6_80, 30),
    (False, EALGFI, CLASS_6_100, 32),
    (False, EALGFI, CLASS_7_20, 55),
    (False, EALGFI, CLASS_7_40, 111),
    (False, EALGFI, CLASS_7_60, 159),
    (False, EALGFI, CLASS_7_80, 232),
    (False, EALGFI, CLASS_7_100, 271),
    (False, EALGFI, CLASS_8_20, 58),
    (False, EALGFI, CLASS_8_40, 113),
    (False, EALGFI, CLASS_8_60, 161),
    (False, EALGFI, CLASS_8_80, 224),
    (False, EALGFI, CLASS_8_100, 277),
    (False, EALGFI, CLASS_9_20, 143),
    (False, EALGFI, CLASS_9_40, 278),
    (False, EALGFI, CLASS_9_60, 437),
    (False, EALGFI, CLASS_9_80, 577),
    (False, EALGFI, CLASS_9_100, 695),
    (False, EALGFI, CLASS_10_20, 42),
    (False, EALGFI, CLASS_10_40, 74),
    (False, EALGFI, CLASS_10_60, 101),
    (False, EALGFI, CLASS_10_80, 128),
    (False, EALGFI, CLASS_10_100, 160),
    (False, GRASP_VND, CLASS_1_20, 71),
    (False, GRASP_VND, CLASS_1_40, 134),
    (False, GRASP_VND, CLASS_1_60, 200),
    (False, GRASP_VND, CLASS_1_80, 275),
    (False, GRASP_VND, CLASS_1_100, 317),
    (False, GRASP_VND, CLASS_2_20, 10),
    (False, GRASP_VND, CLASS_2_40, 19),
    (False, GRASP_VND, CLASS_2_60, 25),
    (False, GRASP_VND, CLASS_2_80, 31),
    (False, GRASP_VND, CLASS_2_100, 39),
    (False, GRASP_VND, CLASS_3_20, 51),
    (False, GRASP_VND, CLASS_3_40, 94),
    (False, GRASP_VND, CLASS_3_60, 139),
    (False, GRASP_VND, CLASS_3_80, 189),
    (False, GRASP_VND, CLASS_3_100, 223),
    (False, GRASP_VND, CLASS_4_20, 10),
    (False, GRASP_VND, CLASS_4_40, 19),
    (False, GRASP_VND, CLASS_4_60, 25),
    (False, GRASP_VND, CLASS_4_80, 31),
    (False, GRASP_VND, CLASS_4_100, 38),
    (False, GRASP_VND, CLASS_5_20, 65),
    (False, GRASP_VND, CLASS_5_40, 119),
    (False, GRASP_VND, CLASS_5_60, 180),
    (False, GRASP_VND, CLASS_5_80, 247),
    (False, GRASP_VND, CLASS_5_100, 282),
    (False, GRASP_VND, CLASS_6_20, 10),
    (False, GRASP_VND, CLASS_6_40, 17),
    (False, GRASP_VND, CLASS_6_60, 21),
    (False, GRASP_VND, CLASS_6_80, 30),
    (False, GRASP_VND, CLASS_6_100, 34),
    (False, GRASP_VND, CLASS_7_20, 55),
    (False, GRASP_VND, CLASS_7_40, 111),
    (False, GRASP_VND, CLASS_7_60, 159),
    (False, GRASP_VND, CLASS_7_80, 232),
    (False, GRASP_VND, CLASS_7_100, 271),
    (False, GRASP_VND, CLASS_8_20, 58),
    (False, GRASP_VND, CLASS_8_40, 113),
    (False, GRASP_VND, CLASS_8_60, 161),
    (False, GRASP_VND, CLASS_8_80, 224),
    (False, GRASP_VND, CLASS_8_100, 278),
    (False, GRASP_VND, CLASS_9_20, 143),
    (False, GRASP_VND, CLASS_9_40, 278),
    (False, GRASP_VND, CLASS_9_60, 437),
    (False, GRASP_VND, CLASS_9_80, 577),
    (False, GRASP_VND, CLASS_9_100, 695),
    (False, GRASP_VND, CLASS_10_20, 42),
    (False, GRASP_VND, CLASS_10_40, 74),
    (False, GRASP_VND, CLASS_10_60, 100),
    (False, GRASP_VND, CLASS_10_80, 129),
    (False, GRASP_VND, CLASS_10_100, 159),
    (False, GRASP_VND, BENG_1_8, 54),
    (False, GRASP_VND, BENG_9_10, 13),
)


def get_related_work(
        with_rotation: bool | None = None,
        without_rotation: bool | None = None,
        algo_select: Callable[[Element], bool] =
        lambda _: True,
        inst_group_select: Callable[[Element], bool] =
        lambda _: True) -> tuple[
        tuple[bool, Element, Element, int], ...]:
    """
    Get the related work of a given type.

    :param with_rotation: include the data with rotation
    :param without_rotation: include the data without rotation
    :param algo_select: the algorithm selector
    :param inst_group_select: the instance group selector
    :return: An iterable with the related works
    """
    res: Iterable[tuple[bool, Element, Element, int]]
    if (with_rotation is None) and (without_rotation is None):
        res = __RW__AVERAGE
    else:
        if with_rotation is None:
            with_rotation = not without_rotation
        if without_rotation is None:
            without_rotation = not with_rotation

        if without_rotation and with_rotation:
            res = __RW__AVERAGE
        elif with_rotation:
            res = (x for x in __RW__AVERAGE if x[0])
        elif without_rotation:
            res = (x for x in __RW__AVERAGE if not x[0])
        else:
            return ()
    return tuple(sorted(filter(
        lambda x: algo_select(x[1]) and inst_group_select(x[2]), res),
        key=lambda v: (v[0], INST_GROUP_SORT_KEY(v[2]), v[1], v[3])))


def make_comparison_table_data(
        data: Iterable[PackingResult],
        with_rotation: bool,
        algo_from_pr: Callable[[PackingResult], Element] = lambda x: Element(
            x.end_result.algorithm, x.end_result.objective),
        algo_sort_key: Callable[[Element], Any] = lambda x: (
        -1 if (x.reference == NO_REF) else 1, x),
        rw_algo_selector: Callable[[Element], bool] = lambda _: True,
        aggregator: Callable[[Iterable[int | float]], int | float] = mean,
        additional: Callable[[Element], Iterable[tuple[
            Element, Callable[[Iterable[int | float]], int | float]]]] =
        lambda _: ()) -> tuple[tuple[Element, ...], tuple[tuple[
        Element, tuple[int | float | None, ...]], ...]]:
    """
    Create the data for an end result comparison table.

    :param data: the source data
    :param with_rotation: are we doing stuff with rotation?
    :param algo_from_pr: convert a packing result to an algorithm
    :param algo_sort_key: the algorithm sort key
    :param rw_algo_selector: the related work algorithm selector
    :param aggregator: the routine for per-instance averaging of bins
    :param additional: an additional column constructor

    :return: the table data: the title row columns followed by the data
        row-by-row, each row leading with an instance group identifier
    """
    per_algo_data: dict[Element, tuple[Callable[[
        Iterable[int | float]], int | float], dict[str, list[int]]]] = {}
    for pr in data:
        algo: Element = algo_from_pr(pr)
        inst_dict: dict[str, list[int]]
        if algo in per_algo_data:
            inst_dict = per_algo_data[algo][1]
        else:
            inst_dict = {}
            per_algo_data[algo] = aggregator, inst_dict
        inst: str = pr.end_result.instance
        val: int = check_int_range(
            pr.objectives[BIN_COUNT_NAME], BIN_COUNT_NAME, 1, 1_000_000_000)
        if inst in inst_dict:
            inst_dict[inst].append(val)
        else:
            inst_dict[inst] = [val]

    groups: Final[list[Element]] = sorted({
        INSTANCES_TO_GROUPS[k] for kk in per_algo_data.values()
        for k in kk[1]}, key=INST_GROUP_SORT_KEY)

    # get the averages
    algo_per_group_data: Final[dict[Element, dict[Element, int | float]]] = {}
    for algo, inst_dict_and_avg in per_algo_data.items():
        dothis: list[tuple[Element, Callable[[
            Iterable[int | float]], int | float]]] = [
            (algo, inst_dict_and_avg[0])]
        dothis.extend(additional(algo))

        for xyz in dothis:
            avg: Callable[[Iterable[int | float]], int | float] = xyz[1]
            inst_dict = inst_dict_and_avg[1]
            ag: dict[Element, int | float] = {}
            algo_per_group_data[xyz[0]] = ag
            algo_runs: int | None = None
            for g in groups:
                gis: tuple[str, ...] = GROUPS_TO_INSTANCES[g]
                gisd: list[list[int | float]] = []
                ok: bool = True
                for inst in gis:
                    if inst not in inst_dict:
                        ok = False
                    gisd.append(cast(list[int | float], inst_dict[inst]))
                if ok:
                    if list.__len__(gisd) <= 0:
                        raise ValueError(f"Empty group {g!r} for {algo!r}?")
                    if list.__len__(gisd) != tuple.__len__(gis):
                        raise ValueError(
                            f"Some data missing in {g!r} for {algo!r}?")
                    ll = list.__len__(gisd[0])
                    if algo_runs is None:
                        algo_runs = ll
                    elif algo_runs != ll:
                        raise ValueError(
                            f"Inconsistent number of runs for {algo!r}: "
                            f"{ll} in {gis[0]!r} vs. {algo_runs}.")
                    if not all(list.__len__(xxx) == ll for xxx in gisd):
                        raise ValueError(
                            f"Inconsistent run numbers in {g!r} for {algo!r}.")
                    ag[g] = int(round(sum(cast(Iterable, (
                        try_int(avg(v)) for v in gisd)))))
                elif len(gisd) > 0:
                    raise ValueError(f"Incomplete group {g!r} for {algo!r}.")

    rw: tuple[tuple[bool, Element, Element, float], ...] = get_related_work(
        with_rotation=with_rotation, without_rotation=not with_rotation,
        algo_select=rw_algo_selector, inst_group_select=groups.__contains__)

    algorithms: list[Element] = list({rww[1] for rww in rw})
    algorithms.sort(key=algo_sort_key)
    our_algorithms = sorted(algo_per_group_data.keys(), key=algo_sort_key)

    rows: list[tuple[Element, tuple[int | float | None, ...]]] = []
    for g in groups:
        xdata: list[int | float | None] = []
        for algo in algorithms:
            found: bool = False
            for zz in rw:
                if zz[1] != algo:
                    continue
                if zz[2] != g:
                    continue
                found = True
                xdata.append(zz[3])
                break
            if found:
                continue
            xdata.append(None)
        for algo in our_algorithms:
            agz: dict[Element, int | float] = algo_per_group_data[algo]
            if g not in agz:
                xdata.append(None)
                continue
            xdata.append(agz[g])
        rows.append((g, tuple(xdata)))

    algorithms.extend(our_algorithms)
    return tuple(algorithms), tuple(rows)


def make_comparison_table(
        dest: Callable[[str], Any],
        data: tuple[tuple[Element, ...], tuple[tuple[Element, tuple[
            int | float | None, ...]], ...]],
        name_to_strs: Callable[[Element], tuple[str, str]] = lambda s: (
        s.name.split("_")[0], s.name_suffix),
        format_best: Callable[[str], str] =
        lambda s: f"\\textbf{{{s}}}",
        count_best_over: Iterable[tuple[Iterable[Element], str]] = ()) -> None:
    """
    Make a comparison table in LaTeX.

    :param dest: the destination
    :param data: the source data
    :param name_to_strs: the converter of names to strings
    :param format_best: the formatter for best values
    :param count_best_over: a set of instances to include in the best counting
        and a corresponding label
    """
    inst_names: Final[list[tuple[str, str]]] = [
        name_to_strs(d[0]) for d in data[1]]
    inst_cols: int = 2 if any(
        str.__len__(n[1]) > 0 for n in inst_names) else 1

    dest("% begin auto-generated LaTeX table comparing "
         "end results with related work.%")
    writer: list[str] = [r"\begin{tabular}{@{}r"]
    if inst_cols > 1:
        writer.append("@{}l")
    for _ in range(tuple.__len__(data[0])):
        writer.append("r")
    writer.append("@{}}%")
    dest("".join(writer))
    writer.clear()
    dest(r"\hline%")

    head_names: Final[list[tuple[str, str]]] = [
        name_to_strs(d) for d in data[0]]
    head_names.insert(0, ("instance", "group"))
    if inst_cols > 1:
        head_names.insert(0, ("instance", "group"))
    head_count: Final[int] = list.__len__(head_names)

    for dim in (0, 1):
        i = 0
        while i < head_count:
            head = head_names[i]
            next_i = i + 1
            while (next_i < head_count) and (
                    head_names[next_i][dim] == head[dim]):
                next_i += 1

            col_txt: str = f"\\multicolumn{{{next_i - i}}}{{"
            if i <= 0:
                col_txt = f"{col_txt}@{{}}"
            col_txt = f"{col_txt}c"
            if next_i >= head_count:
                col_txt = f"{col_txt}@{{}}"
            col_txt = f"{col_txt}}}{{{head[dim]}}}"
            if next_i >= head_count:
                col_txt = f"{col_txt}\\\\%"
            writer.append(col_txt)
            i = next_i
        dest("&".join(writer))
        writer.clear()
    last_first_name = ""

    count_best: list[tuple[str, list[int]]] = []
    count_as: dict[Element, list[list[int]]] = {}
    for count_this, title in count_best_over:
        lst: list[int] = [0] * tuple.__len__(data[0])
        count_best.append((title, lst))
        for k in count_this:
            if k in count_as:
                count_as[k].append(lst)
            else:
                count_as[k] = [lst]

    for i, data_row in enumerate(data[1]):
        inst_name: tuple[str, str] = inst_names[i]
        if inst_name[0] != last_first_name:
            last_first_name = inst_name[0]
            dest(r"\hline%")
        writer.append(last_first_name)
        if inst_cols > 1:
            in1: str = str.strip(inst_name[1])
            if str.__len__(in1) > 0:
                in1 = str.strip(f"/{in1}")
                if str.__len__(in1) <= 1:
                    in1 = ""
            writer.append(in1)

        row_data: tuple[int | float | None, ...] = data_row[1]
        minimum: int | float = inf
        for ddd in row_data:
            if (ddd is not None) and (ddd < minimum):
                minimum = ddd
        if not isfinite(minimum):
            raise ValueError(f"Huuhhh? {inst_name} has non-finite min?")
        instance: Element = data_row[0]
        for iij, ddd in enumerate(row_data):
            if ddd is None:
                writer.append("")
                continue
            writer.append(
                format_best(str(ddd)) if ddd <= minimum else str(ddd))
            if (ddd <= minimum) and (instance in count_as):
                for lst in count_as[instance]:
                    lst[iij] += 1
        dest(f"{'&'.join(writer)}\\\\%")
        writer.clear()

    dest(r"\hline%")
    if list.__len__(count_best) > 0:
        for brow in count_best:
            writer.append(
                f"\\multicolumn{{{inst_cols}}}{{@{{}}r}}{{{brow[0]}}}")
            writer.extend(map(str, brow[1]))
            writer[-1] = f"{writer[-1]}\\\\%"
            dest("&".join(writer))
        dest(r"\hline%")
    dest(r"\end{tabular}%")
    dest("% end auto-generated LaTeX table comparing "
         "end results with related work.%")



"""Some examples for distance metrics."""

from typing import Final

import numba  # type: ignore
import numpy as np
from moptipy.utils.nputils import DEFAULT_BOOL


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def swap_distance(p1: np.ndarray, p2: np.ndarray) -> int:
    """
    Compute the swap distance between two permutations `p1` and `p1`.

    This is the minimum number of swaps required to translate `p1` to `p2` and
    vice versa. This function is symmatric.

    An upper bound for the number of maximum number of swaps that could be
    required is the length of the permutation. This upper bound can be derived
    from Selection Sort. Imagine that I want to translate the array `p1` to
    `p2`. I go through `p1` from beginning to end. If, at index `i`, I find
    the right element (`p1[i] == p2[i]`), then I do nothing. If not, then the
    right element must come at some index `j>i` (because all elements before I
    already have fixed). So I swap `p1[i]` with `p1[j]`. Now `p1[i] == p2[i]`
    and I increment `i`. Once I arrive at the end of `p1`, it must hold that
    `all(p1[i] == p2[i])`. At the same time, I have performed at most one swap
    at each index during the iteration. Hence, I can never need more swaps
    than the arrays are long.

    :param p1: the first permutation
    :param p2: the second permutation
    :return: the swap distance, always between `0` and `len(p1)`

    >>> swap_distance(np.array([0, 1, 2, 3]), np.array([3, 1, 2, 0]))
    1
    >>> swap_distance(np.array([0, 1, 2]), np.array([0, 1, 2]))
    0
    >>> swap_distance(np.array([1, 0, 2]), np.array([0, 1, 2]))
    1
    >>> swap_distance(np.array([0, 1, 2]), np.array([1, 0, 2]))
    1
    >>> swap_distance(np.array([0, 1, 2]), np.array([2, 0, 1]))
    2
    >>> swap_distance(np.array([2, 0, 1]), np.array([0, 1, 2]))
    2
    >>> swap_distance(np.arange(10), np.array([4, 8, 1, 5, 9, 3, 6, 0, 7, 2]))
    7
    >>> swap_distance(np.array([4, 8, 1, 5, 9, 3, 6, 0, 7, 2]), np.arange(10))
    7
    """
    n: Final[int] = len(p1)
    x: np.ndarray = p2[np.argsort(p1)]
    unchecked: np.ndarray = np.ones(n, DEFAULT_BOOL)
    result: int = 0

    for i in range(n):
        if unchecked[i]:
            result += 1
            unchecked[i] = False
            j = x[i]
            while j != i:
                unchecked[j] = False
                j = x[j]

    return n - result



"""Perform tests on the Traveling Salesperson Problem."""

from time import monotonic_ns
from typing import Callable, Final, Iterable

import numpy as np
from moptipy.api.algorithm import Algorithm
from moptipy.api.objective import Objective
from moptipy.spaces.permutations import Permutations
from moptipy.tests.algorithm import validate_algorithm
from moptipy.tests.objective import validate_objective
from numpy.random import Generator, default_rng
from pycommons.types import type_error

from moptipyapps.tsp.instance import Instance
from moptipyapps.tsp.tour_length import TourLength

#: the internal random number generator
__RANDOM: Final[Generator] = default_rng()


def tsp_instances_for_tests(
        random: Generator = __RANDOM,
        symmetric: bool = True,
        asymmetric: bool = True) -> Iterable[str]:
    """
    Get a sequence of TSP instances to test on.

    :param random: the random number generator to use
    :param symmetric: include symmetric instances
    :param asymmetric: include asymmetric instances
    :returns: an iterable of TSP instance names
    """
    if not isinstance(symmetric, bool):
        raise type_error(symmetric, "symmetric", bool)
    if not isinstance(asymmetric, bool):
        raise type_error(asymmetric, "asymmetric", bool)

    instances: tuple[str, ...]
    if symmetric and asymmetric:
        instances = Instance.list_resources(True, True)
    elif symmetric:
        instances = Instance.list_resources(True, False)
    elif asymmetric:
        instances = Instance.list_resources(False, True)
    else:
        raise ValueError(
            "at least of one symmetric or asymmetric must be TRUE")
    use_insts = list(instances)
    while len(use_insts) > 20:
        del use_insts[random.integers(len(use_insts))]
    random.shuffle(use_insts)
    return use_insts


def make_tour_valid(random: Generator, y: np.ndarray) -> np.ndarray:
    """
    Create valid TSP tours.

    :param random: the random number generator to use
    :param y: the input tour
    :returns: the valid version of `y`
    """
    y[0:len(y)] = range(len(y))
    random.shuffle(y)
    return y


def make_tour_invalid(random: Generator, y: np.ndarray) -> np.ndarray:
    """
    Create invalid tours.

    :param random: the random number generator to use
    :param y: the input tour
    :returns: the invalid version of `y`
    """
    ly: Final[int] = len(y)
    y[0:ly] = range(ly)
    random.shuffle(y)
    yorig = np.copy(y)

    ri = random.integers
    end_time: Final[int] = monotonic_ns() + 20_000_000_000
    while np.all(y == yorig):
        if monotonic_ns() >= end_time:
            y[0] = y[1]
            return y
        if ri(2) <= 0:
            z1 = z2 = ri(ly)
            while z1 == z2:
                if monotonic_ns() >= end_time:
                    y[0] = y[1]
                    return y
                z2 = ri(ly)
            y[z1] = y[z2]
        if ri(2) <= 0:
            y[ri(ly)] = ri(ly, 10 * ly)
        if ri(2) <= 0:
            y[ri(ly)] = ri(-2 * ly, -1)
    return y


def validate_algorithm_on_1_tsp(
        algorithm: Algorithm | Callable[
            [Instance, Permutations], Algorithm],
        instance: str | None = None, max_fes: int = 256,
        random: Generator = __RANDOM) -> None:
    """
    Check the validity of a black-box algorithm on one TSP instance.

    :param algorithm: the algorithm or algorithm factory
    :param instance: the instance name, or `None` to randomly pick one
    :param max_fes: the maximum number of FEs
    :param random: the default random generator to use
    """
    if not (isinstance(algorithm, Algorithm) or callable(algorithm)):
        raise type_error(algorithm, "algorithm", Algorithm, True)
    if instance is None:
        instance = str(random.choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "TSP instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, f"loaded bin TSP instance {instance!r}",
                         Instance)

    search_space = Permutations.standard(inst.n_cities)
    objective = TourLength(inst)
    if callable(algorithm):
        algorithm = algorithm(inst, search_space)
    if not isinstance(algorithm, Algorithm):
        raise type_error(algorithm, "algorithm", Algorithm, call=True)

    validate_algorithm(algorithm=algorithm,
                       solution_space=search_space,
                       objective=objective,
                       max_fes=max_fes)


def validate_algorithm_on_tsp(
        algorithm: Callable[[Instance, Permutations], Algorithm],
        symmetric: bool = True, asymmetric: bool = True,
        max_fes: int = 256, random: Generator = __RANDOM) -> None:
    """
    Validate an algorithm on a set of TSP instances.

    :param algorithm: the algorithm factory
    :param symmetric: include symmetric instances
    :param asymmetric: include asymmetric instances
    :param max_fes: the maximum FEs
    :param random: the random number generator
    """
    end_time: Final[int] = monotonic_ns() + 20_000_000_000
    for i in tsp_instances_for_tests(random, symmetric, asymmetric):
        if monotonic_ns() >= end_time:
            return
        validate_algorithm_on_1_tsp(algorithm, i, max_fes, random)


def validate_objective_on_1_tsp(
        objective: Objective | Callable[[Instance], Objective],
        instance: str | None = None,
        random: Generator = __RANDOM) -> None:
    """
    Validate an objective function on 1 TSP instance.

    :param objective: the objective function or a factory creating it
    :param instance: the instance name
    :param random: the random number generator
    """
    if instance is None:
        instance = str(random.choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "TSP instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, f"loaded TSP instance {instance!r}",
                         Instance)

    if callable(objective):
        objective = objective(inst)

    validate_objective(
        objective=objective,
        solution_space=Permutations.standard(inst.n_cities),
        make_solution_space_element_valid=make_tour_valid,
        is_deterministic=True)


def validate_objective_on_tsp(
        objective: Objective | Callable[[Instance], Objective],
        symmetric: bool = True, asymmetric: bool = True,
        random: Generator = __RANDOM) -> None:
    """
    Validate an objective function on TSP instances.

    :param objective: the objective function or a factory creating it
    :param symmetric: include symmetric instances
    :param asymmetric: include asymmetric instances
    :param random: the random number generator
    """
    end_time: Final[int] = monotonic_ns() + 20_000_000_000
    for i in tsp_instances_for_tests(random, symmetric, asymmetric):
        if monotonic_ns() >= end_time:
            return
        validate_objective_on_1_tsp(objective, i, random)



"""
An instance of the Quadratic Assignment Problem.

In this module, we provide the class :class:`~Instance` that encapsulates all
information of a problem instance of the Quadratic Assignment Problem (QAP).
The QAP aims to locate facilities at locations such that the flow-distance
product sum combining a flows of goods between instances with the distances of
the locations becomes minimal. Each instance therefore presents a matrix with
:attr:`~moptipyapps.qap.instance.Instance.distances` and a matrix with flows
:attr:`~moptipyapps.qap.instance.Instance.flows`.

1. Eliane Maria Loiola, Nair Maria Maia de Abreu, Paulo Oswaldo
   Boaventura-Netto, Peter Hahn, and Tania Querido. A survey for the
   Quadratic Assignment Problem. European Journal of Operational Research.
   176(2):657-690. January 2007. https://doi.org/10.1016/j.ejor.2005.09.032.
2. Rainer E. Burkard, Eranda Çela, Panos M. Pardalos, and
   Leonidas S. Pitsoulis. The Quadratic Assignment Problem. In Ding-Zhu Du,
   Panos M. Pardalos, eds., Handbook of Combinatorial Optimization,
   pages 1713-1809, 1998, Springer New York, NY, USA.
   https://doi.org/10.1007/978-1-4613-0303-9_27.


We additionally provide access to several standard QAP benchmark instances
via the :meth:`~Instance.from_resource` and :meth:`~Instance.list_resources`
methods. The standard benchmark instances stem from QAPLIB, a library of QAP
instances, which can be found at <https://qaplib.mgi.polymtl.ca> and
<https://coral.ise.lehigh.edu/data-sets/qaplib>.

1. QAPLIB - A Quadratic Assignment Problem Library. The Websites
   <https://qaplib.mgi.polymtl.ca/> (updated 2018) and
   <https://coral.ise.lehigh.edu/data-sets/qaplib/> (updated 2011), including
   the benchmark instances, on visited 2023-10-21.
2. Rainer E. Burkard, Stefan E. Karisch, and Franz Rendl. QAPLIB - A Quadratic
   Assignment Problem Library. Journal of Global Optimization. 10:391-403,
   1997. https://doi.org/10.1023/A:1008293323270.
"""


from typing import Final, Iterable, cast

import numba  # type: ignore
import numpy as np
from moptipy.api.component import Component
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import (
    DEFAULT_UNSIGNED_INT,
    int_range_to_dtype,
    is_np_int,
)
from moptipy.utils.strings import sanitize_name
from pycommons.types import check_int_range, check_to_int_range, type_error

from moptipyapps.qap.qaplib import open_resource_stream


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def trivial_bounds(distances: np.ndarray, flows: np.ndarray) \
        -> tuple[int, int]:
    """
    Compute the trivial bounds for the QAP objective.

    A trivial upper bound is to multiply the largest flow with the largest
    distance, the second-largest flow with the second-largest distance, the
    third-largest flow with the third-largest distance, and so on.

    A trivial lower bound is to multiply the largest flow with the shortest
    distance, the second-largest flow with the second-shortest distance, the
    third-largest flow with the third-shortest distance, and so on.

    :param distances: the distance matrix
    :param flows: the flow matrix
    :return: the lower and upper bounds

    >>> dst = np.array([[0, 1, 2], [3, 0, 4], [5, 6, 0]])
    >>> flws = np.array([[0, 95, 86], [23, 0, 55], [24, 43, 0]])
    >>> 0*95 + 0*86 + 0*55 + 1*43 + 2*24 + 3*23 + 4*0 + 5*0 + 6*0
    160
    >>> 6*95 + 5*86 + 4*55 + 3*43 + 2*24 + 1*23 + 0*0 + 0*0 + 0*0
    1420
    >>> trivial_bounds(dst, flws)
    (160, 1420)
    """
    n: int = len(distances)
    n *= n
    df_ub: Final[np.ndarray] = np.empty(n, DEFAULT_UNSIGNED_INT)
    df_ub[:] = distances.flatten()
    df_ub.sort()
    df_lb: Final[np.ndarray] = np.empty(n, DEFAULT_UNSIGNED_INT)
    df_lb[:] = df_ub[::-1]
    ff: Final[np.ndarray] = np.empty(n, DEFAULT_UNSIGNED_INT)
    ff[:] = flows.flatten()
    ff.sort()
    return (int(np.multiply(df_lb, ff, df_lb).sum()),
            int(np.multiply(df_ub, ff, df_ub).sum()))


def _flow_or_dist_to_int(val: str) -> int:
    """
    Convert a flow or distance string to an integer.

    :param val: the value
    :return: the integer
    """
    return check_to_int_range(val, "value", 0, 1_000_000_000_000_000)


#: the instances of the QAPLib library
_INSTANCES: Final[tuple[str, ...]] = (
    "bur26a", "bur26b", "bur26c", "bur26d", "bur26e", "bur26f", "bur26g",
    "bur26h", "chr12a", "chr12b", "chr12c", "chr15a", "chr15b", "chr15c",
    "chr18a", "chr18b", "chr20a", "chr20b", "chr20c", "chr22a", "chr22b",
    "chr25a", "els19", "esc16a", "esc16b", "esc16c", "esc16d", "esc16e",
    "esc16f", "esc16g", "esc16h", "esc16i", "esc16j", "esc32a", "esc32b",
    "esc32c", "esc32d", "esc32e", "esc32g", "esc32h", "esc64a", "esc128",
    "had12", "had14", "had16", "had18", "had20", "kra30a", "kra30b", "kra32",
    "lipa20a", "lipa20b", "lipa30a", "lipa30b", "lipa40a", "lipa40b",
    "lipa50a", "lipa50b", "lipa60a", "lipa60b", "lipa70a", "lipa70b",
    "lipa80a", "lipa80b", "lipa90a", "lipa90b", "nug12", "nug14", "nug15",
    "nug16a", "nug16b", "nug17", "nug18", "nug20", "nug21", "nug22", "nug24",
    "nug25", "nug27", "nug28", "nug30", "rou12", "rou15", "rou20", "scr12",
    "scr15", "scr20", "sko42", "sko49", "sko56", "sko64", "sko72", "sko81",
    "sko90", "sko100a", "sko100b", "sko100c", "sko100d", "sko100e", "sko100f",
    "ste36a", "ste36b", "ste36c", "tai12a", "tai12b", "tai15a", "tai15b",
    "tai17a", "tai20a", "tai20b", "tai25a", "tai25b", "tai30a", "tai30b",
    "tai35a", "tai35b", "tai40a", "tai40b", "tai50a", "tai50b", "tai60a",
    "tai60b", "tai64c", "tai80a", "tai80b", "tai100a", "tai100b", "tai150b",
    "tai256c", "tho30", "tho40", "tho150", "wil50", "wil100")

#: the lower bounds provided at <https://qaplib.mgi.polymtl.ca/>
_BOUNDS: Final[dict[str, int]] = {
    "nug14": 1014, "nug15": 1150, "nug16a": 1610, "nug16b": 1240,
    "nug17": 1732, "nug18": 1930, "nug20": 2570, "nug21": 2438, "nug22": 3596,
    "nug24": 3488, "nug25": 3744, "nug27": 5234, "nug28": 5166, "nug30": 6124,
    "rou12": 235528, "rou15": 354210, "rou20": 725522, "scr12": 31410,
    "scr15": 51140, "scr20": 110030, "sko100a": 143846, "sko100b": 145522,
    "sko100c": 139881, "sko100d": 141289, "sko100e": 140893, "sko100f": 140691,
    "sko42": 15332, "sko49": 22650, "sko56": 33385, "sko64": 47017,
    "sko72": 64455, "sko81": 88359, "sko90": 112423, "ste36a": 9526,
    "ste36b": 15852, "ste36c": 8239110, "tai100a": 17853840,
    "tai100b": 1151591000, "tai12a": 224416, "tai12b": 39464925,
    "tai150b": 441786736, "tai15a": 388214, "tai15b": 51765268,
    "tai17a": 491812, "tai20a": 703482, "tai20b": 122455319,
    "tai256c": 44095032, "tai25a": 1167256, "tai25b": 344355646,
    "tai30a": 1706855, "tai30b": 637117113, "tai35a": 2216627,
    "tai35b": 269532400, "tai40a": 2843274, "tai40b": 608808400,
    "tai50a": 4390920, "tai50b": 431090700, "tai60a": 6325978,
    "tai60b": 592371800, "tai64c": 1855928, "tai80a": 11657010,
    "tai80b": 786298800, "tho150": 7620628, "tho30": 149936, "tho40": 226490,
    "wil100": 268955, "wil50": 48121, "bur26a": 5426670, "bur26b": 3817852,
    "bur26c": 5426795, "bur26d": 3821225, "bur26e": 5386879, "bur26f": 3782044,
    "bur26g": 10117172, "bur26h": 7098658, "chr12a": 9552, "chr12b": 9742,
    "chr12c": 11156, "chr15a": 9896, "chr15b": 7990, "chr15c": 9504,
    "chr18a": 11098, "chr18b": 1534, "chr20a": 2192, "chr20b": 2298,
    "chr20c": 14142, "chr22a": 6156, "chr22b": 6194, "chr25a": 3796,
    "els19": 17212548, "esc128": 64, "esc16a": 68, "esc16b": 292,
    "esc16c": 160, "esc16d": 16, "esc16e": 28, "esc16f": 0, "esc16g": 26,
    "esc16h": 996, "esc16i": 14, "esc16j": 8, "esc32a": 130, "esc32b": 168,
    "esc32c": 642, "esc32d": 200, "esc32e": 2, "esc32g": 6, "esc32h": 438,
    "esc64a": 116, "had12": 1652, "had14": 2724, "had16": 3720, "had18": 5358,
    "had20": 6922, "kra30a": 88900, "kra30b": 91420, "kra32": 88700,
    "lipa20a": 3683, "lipa20b": 27076, "lipa30a": 13178, "lipa30b": 151426,
    "lipa40a": 31538, "lipa40b": 476581, "lipa50a": 62093, "lipa50b": 1210244,
    "lipa60a": 107218, "lipa60b": 2520135, "lipa70a": 169755,
    "lipa70b": 4603200, "lipa80a": 253195, "lipa80b": 7763962,
    "lipa90a": 360630, "lipa90b": 12490441, "nug12": 578}

#: The best-known solutions of the QAPLIB instances that are not yet solved to
#: optimality, as of 2024-05-09 on https://qaplib.mgi.polymtl.ca/
_BKS: Final[dict[str, tuple[str, int]]] = {
    "sko90": ("T1991RTSFTQAP,T1995COISFTQAP", 115534),
    "sko100a": ("FF1993GHFTQAP", 152002),
    "sko100b": ("FF1993GHFTQAP", 153890),
    "sko100c": ("FF1993GHFTQAP", 147862),
    "sko100d": ("FF1993GHFTQAP", 149576),
    "sko100e": ("FF1993GHFTQAP", 149150),
    "sko100f": ("FF1993GHFTQAP", 149036),
    "tai30a": ("T1991RTSFTQAP,T1995COISFTQAP", 1818146),
    "tai35a": ("T1991RTSFTQAP,T1995COISFTQAP", 2422002),
    "tai35b": ("T1991RTSFTQAP,T1995COISFTQAP", 283315445),
    "tai40a": ("T1991RTSFTQAP,T1995COISFTQAP", 3139370),
    "tai40b": ("T1991RTSFTQAP,T1995COISFTQAP", 637250948),
    "tai50a": ("M2008AIOTITSAFTQAP", 4938796),
    "tai50b": ("T1991RTSFTQAP,T1995COISFTQAP", 458821517),
    "tai60a": ("M2005ATSAFTQAP", 7205962),
    "tai60b": ("T1991RTSFTQAP,T1995COISFTQAP", 608215054),
    "tai80a": ("M2008AIOTITSAFTQAP", 13499184),
    "tai80b": ("T1991RTSFTQAP,T1995COISFTQAP", 818415043),
    "tai100a": ("M2008AIOTITSAFTQAP", 21044752),
    "tai100b": ("T1991RTSFTQAP,T1995COISFTQAP", 1185996137),
    "tai150b": ("TG1997AMFTQAP", 498896643),
    "tai256c": ("S1997MMASFQAP", 44759294),
    "tho40": ("TB1994AISAAFTQAP", 240516),
    "tho150": ("M2003AMSAAFTQAP", 8133398),
    "wil100": ("FF1993GHFTQAP", 273038),
}


class Instance(Component):
    """An instance of the Quadratic Assignment Problem."""

    def __init__(self, distances: np.ndarray, flows: np.ndarray,
                 lower_bound: int | None = None,
                 upper_bound: int | None = None,
                 name: str | None = None) -> None:
        """
        Create an instance of the quadratic assignment problem.

        :param distances: the distance matrix
        :param flows: the flow matrix
        :param lower_bound: the optional lower bound
        :param upper_bound: the optional upper bound
        :param name: the name of this instance
        """
        super().__init__()
        if not isinstance(distances, np.ndarray):
            raise type_error(distances, "distances", np.ndarray)
        if not isinstance(flows, np.ndarray):
            raise type_error(flows, "flows", np.ndarray)
        shape: tuple[int, ...] = distances.shape
        if len(shape) != 2:
            raise ValueError("distance matrix must have two dimensions, but "
                             f"has {len(shape)}, namely {shape}.")
        if shape[0] != shape[1]:
            raise ValueError(
                f"distance matrix must be square, but has shape {shape}.")
        if not is_np_int(distances.dtype):
            raise ValueError("distance matrix must be integer, but has "
                             f"dtype {distances.dtype}.")
        if shape != flows.shape:
            raise ValueError(
                f"flow matrix has shape {flows.shape} and distance matrix has"
                f" shape {shape}, which are different.")
        if not is_np_int(flows.dtype):
            raise ValueError(
                f"flow matrix must be integer, but has dtype {flows.dtype}.")

        lb, ub = trivial_bounds(distances, flows)
        if lower_bound is not None:
            lb = max(lb, check_int_range(
                lower_bound, "lower_bound", 0, 1_000_000_000_000_000))
        if upper_bound is not None:
            ub = min(ub, check_int_range(
                upper_bound, "upper_bound", 0, 1_000_000_000_000_000))
        if lb > ub:
            raise ValueError(f"lower bound = {lb} > upper_bound = {ub}!")
        dtype: Final[np.dtype] = int_range_to_dtype(min_value=0, max_value=ub)
        #: the scale of the problem
        self.n: Final[int] = shape[0]
        if name is None:
            name = f"qap{self.n}_{lb}_{ub}"
        else:
            sn: Final[str] = sanitize_name(name)
            if name != sn:
                raise ValueError(f"name={name!r} sanitizes to {sn!r}.")
        #: the name of this instance
        self.name: Final[str] = name
        #: the distance matrix
        self.distances: Final[np.ndarray] = \
            distances.astype(dtype) if distances.dtype != dtype else distances
        #: the flows
        self.flows: Final[np.ndarray] = \
            flows.astype(dtype) if flows.dtype != dtype else flows
        #: the lower bound for the QAP objective
        self.lower_bound: Final[int] = lb
        #: the upper bound for the QAP objective
        self.upper_bound: Final[int] = ub

    def __str__(self):
        """
        Get the name of this instance.

        :return: :attr:`~name`
        """
        return self.name

    def bks(self) -> tuple[str, int]:
        """
        Get the best-known solution, if known, the optimum, or lower bound.

        A tuple with a string identifying the source of the value and a value
        corresponding to the best-known solution:

        - `("OPT", xxx)`: the problem instance has been solved to optimality
          and `xxx` is the objective value of the optimum
        - `("LB", xxx)`: neither the optimum nor a best-known solution are
          available for this instance, so we return the lower bound

        The data is based on https://qaplib.mgi.polymtl.ca/, visited on
        2024-05-09. The following sources are included:

        - "FF1993GHFTQAP": Charles Fleurent and Jacques A. Ferland. Genetic
          Hybrids for the Quadratic Assignment Problem. In Panos M. Pardalos
          and Henry Wolkowicz, eds, *Quadratic Assignment and Related
          Problems, Proceedings of a DIMACS Workshop,* May 20-21, 1993.
          pages 173-187. Providence, RI, USA: American Mathematical Society.
        - "M2003AMSAAFTQAP": Alfonsas Misevičius, A Modified Simulated
          Annealing Algorithm for the Quadratic Assignment Problem.
          *Informatica* 14(4):497-514. January 2003.
          https://doi.org/10.15388/Informatica.2003.037.
        - "M2005ATSAFTQAP": Alfonsas Misevičius. A Tabu Search Algorithm for
          the Quadratic Assignment Problem.
          *Computational Optimization and Applications* 30(1):95-111. 2005.
          https://doi.org/10.1007/s10589-005-4562-x.
        - "M2008AIOTITSAFTQAP": Alfonsas Misevičius. An Implementation of
          the Iterated Tabu Search Algorithm for the Quadratic Assignment
          Problem. Working Paper. 2008. Kaunas, Lithuania: Kaunas University
          of Technology.
        - "S1997MMASFQAP": Thomas Stützle. MAX-MIN Ant System for Quadratic
          Assignment Problems. Research Report AIDA-97-04. 1997. Darmstadt,
          Germany: Department of Computer Schience, Darmstadt University of
          Technology.
        - "T1991RTSFTQAP": Éric Taillard. Robust Taboo Search for the
          Quadratic Assignment Problem. *Parallel Computing.*
          17(4-5):443-455. July 1991.
        - "T1995COISFTQAP": Éric D. Taillard. Comparison of Iterative
          Searches for the Quadratic Assignment Problem. *Location Science*
          3(2):87-105. 1995.
        - "TB1994AISAAFTQAP": Ulrich Wilhelm Thonemann and Andreas Bölte.
          An Improved Simulated Annealing Algorithm for the Quadratic
          Assignment Problem. Working Paper 1994. Paderborn, Germany: School
          of Business, Department of Production and Operations Research,
          University of Paderborn.
        - "TG1997AMFTQAP": Éric D. Taillard and Luca-Maria Gambardella.
          Adaptive Memories for the Quadratic Assignment Problem. 1997.
          Technical Report IDSIA-87-97. Lugano, Switzerland: IDSIA.

        :return: a `tuple[str, int]` with the objective value of the
            best-known (if any is available), the optimum, or the lower bound.
        """
        name: Final[str] = self.name
        if name in _BKS:
            return _BKS[name]
        if name in _BOUNDS:
            return "OPT", _BOUNDS[name]
        return "LB", self.lower_bound

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this instance as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("n", self.n)
        logger.key_value("qapLowerBound", self.lower_bound)
        logger.key_value("qapUpperBound", self.upper_bound)

    @staticmethod
    def from_qaplib_stream(stream: Iterable[str],
                           lower_bound: int | None = None,
                           upper_bound: int | None = None,
                           name: str | None = None) -> "Instance":
        """
        Load an instance from a QAPLib-formatted stream.

        :param stream: the stream to load the data from
        :param lower_bound: the optional lower bound
        :param upper_bound: the optional upper bound
        :param name: the name of this instance
        :return: the instance

        >>> ins = Instance.from_qaplib_stream([
        ...         "4", "",
        ...         "1 2 3 4 5 6 7 8 9 10 11 12 13",
        ...         "   14    15   16  ", "",
        ...         "17 18 19 20 21 22 23 24 25 26     27",
        ...         " 28   29 30 31   32"])
        >>> ins.distances
        array([[17, 18, 19, 20],
               [21, 22, 23, 24],
               [25, 26, 27, 28],
               [29, 30, 31, 32]], dtype=int16)
        >>> ins.flows
        array([[ 1,  2,  3,  4],
               [ 5,  6,  7,  8],
               [ 9, 10, 11, 12],
               [13, 14, 15, 16]], dtype=int16)
        >>> ins.lower_bound
        2992
        >>> ins.upper_bound
        3672
        >>> ins.name
        'qap4_2992_3672'
        """
        state: int = 0
        n: int | None = None
        n2: int = -1
        flows: list[int] = []
        dists: list[int] = []
        for oline in stream:
            line = oline.strip()
            if len(line) <= 0:
                continue
            if state == 0:
                n = check_to_int_range(line, "n", 1, 1_000_000)
                n2 = n * n
                state = 1
            else:
                row: Iterable[int] = map(_flow_or_dist_to_int, line.split())
                if state == 1:
                    flows.extend(row)
                    if len(flows) >= n2:
                        state = 2
                        continue
                dists.extend(row)
                if len(dists) >= n2:
                    state = 3
                    break

        if (n is None) or (n <= 0):
            raise ValueError(f"Invalid or unspecified size n={n}.")
        lll: int = len(flows)
        if lll != n2:
            raise ValueError(
                f"Invalid number of flows {lll}, should be n²={n}²={n2}.")
        lll = len(dists)
        if lll != n2:
            raise ValueError(
                f"Invalid number of distances {lll}, should be n²={n}²={n2}.")
        if state != 3:
            raise ValueError(f"Stream is incomplete, state={state}.")

        return Instance(np.array(dists, DEFAULT_UNSIGNED_INT).reshape((n, n)),
                        np.array(flows, DEFAULT_UNSIGNED_INT).reshape((n, n)),
                        lower_bound, upper_bound, name)

    @staticmethod
    def list_resources() -> tuple[str, ...]:
        """
        Get a tuple of all the QAP-lib instances available as resource.

        The original data can be found at <https://qaplib.mgi.polymtl.ca> and
        <https://coral.ise.lehigh.edu/data-sets/qaplib>.

        :return: the tuple with the instance names

        >>> len(Instance.list_resources())
        134
        """
        return _INSTANCES

    @staticmethod
    def from_resource(name: str) -> "Instance":
        """
        Load a QAP-Lib instance from a resource.

        The original data  can be found at <https://qaplib.mgi.polymtl.ca> and
        <https://coral.ise.lehigh.edu/data-sets/qaplib>.

        :param name: the name string
        :return: the instance

        >>> insta = Instance.from_resource("chr25a")
        >>> print(insta.n)
        25
        >>> print(insta.name)
        chr25a
        >>> print(insta.lower_bound)
        3796
        >>> print(insta.upper_bound)
        50474
        """
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        container: Final = Instance.from_resource
        inst_attr: Final[str] = f"__inst_{name}"
        if hasattr(container, inst_attr):  # instance loaded?
            return cast(Instance, getattr(container, inst_attr))

        lb: int | None = _BOUNDS.get(name, None)
        with open_resource_stream(f"{name}.dat") as stream:
            inst: Final[Instance] = Instance.from_qaplib_stream(
                stream, name=name, lower_bound=lb)

        if inst.n <= 1000:
            setattr(container, inst_attr, inst)
        return inst



"""
A set of tools for ordering objects in 1 dimension.

Let's assume that we have `n` objects and a distance metric that can compute
the distance between two objects. We do not know and also do not care about in
how many dimension the objects exist - we just have objects and a distance
metric.

Now we want to find a one-dimensional order of the objects that reflects their
original distance-based topology. For each object `a`, we want that its
closest neighbor in the order is also its actual closest neighbor according to
the distance metric. It's second-closest neighbor should be the actual
second-closest neighbor according to the distance metric. And so on.

Since we only care about the object order and do not want to metrically map
the distances to one dimension, we can represent the solution as permutation
of natural numbers.

Of course, in a one-dimensional order, each object has exactly two closest
neighbors (the one on its left and the one on its right) unless it is situated
either at the beginning or end of the order, in which case it has exactly one
closest neighbor. Based on the actual distance metric, an object may have any
number of closest neighbors, maybe only one, or maybe three equally-far away
objects. So it is not clear whether a perfect mapping to the one-dimensional
permutations even exists. Probably it will not, but this shall not bother us.

Either way, we can try to find one that comes as close as possible to the real
deal.

Another way to describe this problem is as follows:
Imagine that you have `n` objects and only know their mutual distances.
You want to arrange them on a one-dimensional axis in a way that does sort of
reflect their neighborhood structure in whatever space they are originally
located in.

The goal of solving this one-dimensional ordering problem is then to arrange
`n` objects on a 1-dimensional (e.g., horizontal) axis given a distance matrix
describing their location in a potentially high-dimensional or unstructured
space.
The objects should be arranged in such a way that, for each object,

- the nearest neighbors on the 1-dimensional axis are also the nearest
  neighbors in the original space (according to the distance matrix
  provided),
- the second nearest neighbors on the 1-dimensional axis are also the
  second nearest neighbors in the original space (according to the distance
  matrix provided),
- the third nearest neighbors on the 1-dimensional axis are also the third
  nearest neighbors in the original space (according to the distance matrix
  provided),
- and so on; with (e.g., quadratically) decreasing weights of neighbor
  distance ranks.

We do not care about the actual precise distances (e.g., something like
"0.001") between the objects on either the one-dimensional nor the original
space. We only care about the distance ranks, i.e., about "2nd nearest
neighbor," but not "0.012 distance units away." The solutions of this problem
are thus permutations (orders) of the objects. Of course, if we really want
to plot the objects, such a permutation can easily be translated to
`x`-coordinates, say, by dividing the index of an object by the number of
objects, which nets values in `[0,1]`. But basically, we reduce the task to
finding permutations of objects that reflect the neighbor structure of the
original space as closely as possible.

If such a problem is solved correctly, then the arrangement on the
one-dimensional axis should properly reflect the arrangement of the objects in
the original space. Of course, solving this problem exactly may not actually
be possible, since an object on a one-dimensional axis may either have exactly
two `i`-nearest-neighbors (if it is at least `i` slots away from either end of
the permutation) or exactly `1` such neighbor, if it is closer that `i` units.
The object directly at the start of the permutation has only 1 nearest
neighbor (the object that comes next). That next object, however, has two,
namely the first object and the third object. In the original space where the
objects come from, however, there may be any number of "nearest neighbors."
Imagine a two-dimensional space where one object sits at the center of a
circle of other objects. Then all other objects are its nearest neighbors,
whereas an object on the circle either has exactly two nearest neighbors or,
maybe, in the odd situation that the radius equals a multiple of the
distance to the neighbors on the circle, three. Such a structure cannot be
represented exactly in one dimension.

But that's OK.
Because we mainly do this for visualization purposes anyway.

Now here comes the cool thing: We can cast this problem to a Quadratic
Assignment Problem (:mod:`~moptipyapps.qap`). A QAP is defined by a distance
matrix and a flow matrix. The idea is to assign `n` facilities to locations.
The distances between the locations are given by the distance matrix. At the
same time, the flow matrix defines the flows between the facilities. The goal
is to find the assignment of facilities to locations that minimizes the
overall "flow times distance" product sum.

We can translate the "One-Dimensional Ordering Problem" to the QAP as follows:
Imagine that we have `n` unique objects and know the (symmetric) distances
between them.

Let's say that we have four numbers as objects, `1`, `2`, `3`, `4`.
>>> data = [1, 2, 3, 4]
>>> n = len(data)

The distance between two numbers `a` and `b` be `abs(a - b)`.
>>> def dist(a, b):
...     return abs(a - b)

The **original** distance matrix would be

>>> dm = [[dist(i, j) for j in range(n)] for i in range(n)]
>>> print(dm)
[[0, 1, 2, 3], [1, 0, 1, 2], [2, 1, 0, 1], [3, 2, 1, 0]]

Now from this matrix, we can find the nearest neighbors as follows:

>>> from scipy.stats import rankdata  # type: ignore
>>> import numpy as np
>>> rnks = rankdata(dm, axis=1, method="average") - 1
>>> print(rnks)
[[0.  1.  2.  3. ]
 [1.5 0.  1.5 3. ]
 [3.  1.5 0.  1.5]
 [3.  2.  1.  0. ]]

From the perspective of "`1`", "`2`" is the first nearest neighbor,
"`3`" is the second-nearest neighbor, and "`4`" is the third-nearest neighbor.
From the perspective of "`2`", both "`1`" and "`3`" are nearest neighbors, so
they share rank `1.5`.
And so on.

Let's multiply this by `2` to get integers:
[If no fractional ranks appear, then we do not need to multiply by `2`.]

>>> rnks = np.array(rnks * 2, dtype=int)
>>> print(rnks)
[[0 2 4 6]
 [3 0 3 6]
 [6 3 0 3]
 [6 4 2 0]]

Now we want to translate this to a flow matrix (NOT a distance matrix).
What we want is that there is a big flow from each object to its nearest
neighbor, a smaller flow from each object to its second-nearest neighbor,
a yet smaller flow to the third nearest neighbor, and so on.
So all we need to do is to subtract the elements off the diagonal from
`8 = 2*n`.
[If no fractional elements were there, then we would use `n` instead of
`2*n`.]
Anyway, we get:

>>> rnks = np.array([[0 if x == 0 else 2*n - x for x in a]
...                  for a in rnks], int)
>>> print(rnks)
[[0 6 4 2]
 [5 0 5 2]
 [2 5 0 5]
 [2 4 6 0]]

We can see:
The flow from "`1`" to "`2`" is `6`, which higher than the flow from "`1`" to
"`3`" (namely `4`), and so on.
The low from "`2`" to "`1`" is `5`, which is a bit lower than `6` but higher
than `4`, because "`1`" is the "1.5th nearest neighbor of `2`".

Just for good measures, we can weight this qudratically, so we get:

>>> rnks = rnks ** 2
>>> print(rnks)
[[ 0 36 16  4]
 [25  0 25  4]
 [ 4 25  0 25]
 [ 4 16 36  0]]

This way, we give much more importance to "getting the nearest neighbors
right" than getting "far-away neighbors" right. Nice.

OK, but how about the distance matrix for the QAP?
What's the distance?
Well, it's the distance that I would get by arranging the objects in a
specific order.
It is the difference of the indices in the permutation:

>>> print(np.array([[abs(i - j) for j in range(n)] for i in range(n)], int))
[[0 1 2 3]
 [1 0 1 2]
 [2 1 0 1]
 [3 2 1 0]]

This means:
The object that I will place at index `1` has a distance of "`1`" to the
object at index `2`.
It has a distance of "`2`" to the object at index `3`.
It has a distance of "`3`" to the object at index `4`.
The object at index `2` will have a distance of "`1`" to both the objects
at indices `1` and `3` and a distance of "`2`" to the object at index `4`.
And so on.

In other words, objects that are close in their original space have a big
flow between each other.
If we put them at directly neighboring indexes, then we will multiply this
big flow with a small number.
If we put them at distant indices, then we will multiply this big flow with
a large number.

Objects that are far away from each other in their original space have a small
flow between each other.
If we put them at indices that are far apart, we will multiply the larger
index-difference = distance with a small number.
If we put them close, then we multiply a small distance with a small flow -
but we will have to pay for this elsewhere, because then we may need to put
objects with a larger flow between each other farther apart.

>>> from moptipyapps.order1d.instance import Instance
>>> the_instance = Instance.from_sequence_and_distance(
...     [1, 2, 3, 4], dist, 2, 100, ("bla", ), lambda i: str(i))
>>> print(the_instance.flows)
[[ 0 36 16  4]
 [25  0 25  4]
 [ 4 25  0 25]
 [ 4 16 36  0]]
>>> print(the_instance.distances)
[[0 1 2 3]
 [1 0 1 2]
 [2 1 0 1]
 [3 2 1 0]]

There is one little thing that needs to be done:
If we have many objects, then the QAP objective values can get very large.
So it makes sense to define a "horizon" after which the relationship for
objects are ignored.
Above, we chose "100", which had no impact.
If we choose "2", then only the two nearest neighbors would be considered and
we get:

>>> the_instance = Instance.from_sequence_and_distance(
...     [1, 2, 3, 4], dist, 2, 2, ("bla", ), lambda i: str(i))
>>> print(the_instance.flows)
[[ 0 16  4  0]
 [ 9  0  9  0]
 [ 0  9  0  9]
 [ 0  4 16  0]]
>>> print(the_instance.distances)
[[0 1 2 3]
 [1 0 1 2]
 [2 1 0 1]
 [3 2 1 0]]

The distance matrix of the QAP stays the same, but the flow matrix now has
several zeros.
Anything farther away than the second-nearest neighbor will be ignored, i.e.,
get a flow of `0`.
"""



"""An example experiment for bin packing."""

import argparse
from typing import Callable, Final, cast

from moptipy.algorithms.so.fea1plus1 import FEA1plus1
from moptipy.algorithms.so.rls import RLS
from moptipy.api.encoding import Encoding
from moptipy.api.execution import Execution
from moptipy.api.experiment import Parallelism, run_experiment
from moptipy.api.objective import Objective
from moptipy.operators.signed_permutations.op0_shuffle_and_flip import (
    Op0ShuffleAndFlip,
)
from moptipy.operators.signed_permutations.op1_swap_2_or_flip import (
    Op1Swap2OrFlip,
)
from moptipy.spaces.signed_permutations import SignedPermutations
from pycommons.io.path import Path

from moptipyapps.binpacking2d.encodings.ibl_encoding_1 import (
    ImprovedBottomLeftEncoding1,
)
from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.packing_result import DEFAULT_OBJECTIVES
from moptipyapps.binpacking2d.packing_space import PackingSpace
from moptipyapps.shared import moptipyapps_argparser

#: the maximum number of FEs
MAX_FES: Final[int] = 1_000_000


def base_setup(instance: Instance,
               encoding: Callable[[Instance], Encoding],
               objective: Callable[[Instance], Objective]) \
        -> tuple[SignedPermutations, Execution]:
    """
    Create the basic setup.

    :param instance: the instance to use
    :param encoding: the encoding function
    :param objective: the objective function
    :return: the search space and the basic execution
    """
    space: Final[SignedPermutations] = SignedPermutations(
        instance.get_standard_item_sequence())

    return (space, Execution().set_max_fes(MAX_FES).set_log_improvements(True)
            .set_objective(objective(instance))
            .set_encoding(encoding(instance))
            .set_search_space(space)
            .set_solution_space(PackingSpace(instance)))


def rls(instance: Instance,
        encoding: Callable[[Instance], Encoding],
        objective: Callable[[Instance], Objective]) -> Execution:
    """
    Create the RLS setup.

    :param instance: the instance to use
    :param encoding: the encoding function
    :param objective: the objective function
    :return: the RLS execution
    """
    space, execute = base_setup(instance, encoding, objective)
    return execute.set_algorithm(
        RLS(Op0ShuffleAndFlip(space), Op1Swap2OrFlip()))


def fea(instance: Instance,
        encoding: Callable[[Instance], Encoding],
        objective: Callable[[Instance], Objective]) -> Execution:
    """
    Create the FEA setup.

    :param instance: the instance to use
    :param encoding: the encoding function
    :param objective: the objective function
    :return: the RLS execution
    """
    space, execute = base_setup(instance, encoding, objective)
    return execute.set_algorithm(
        FEA1plus1(Op0ShuffleAndFlip(space), Op1Swap2OrFlip()))


def run(base_dir: str, n_runs: int = 23) -> None:
    """
    Run the experiment.

    :param base_dir: the base directory
    :param n_runs: the number of runs, by default `23`
    """
    use_dir: Final[Path] = Path(base_dir)
    use_dir.ensure_dir_exists()

    encodings: Final[tuple[Callable[[Instance], Encoding], ...]] = (
        ImprovedBottomLeftEncoding1,
    )
    instances: list[str] = [
        inst for inst in Instance.list_resources()
        if inst.startswith(("b", "a"))]
    inst_creators: list[Callable[[], Instance]] = [cast(
        Callable[[], Instance], lambda __s=_s: Instance.from_resource(__s))
        for _s in instances]
    namer: Final[Instance] = Instance.from_resource(instances[0])

    for objective in DEFAULT_OBJECTIVES:
        objective_dir: Path = use_dir.resolve_inside(str(objective(namer)))
        objective_dir.ensure_dir_exists()
        for encoding in encodings:
            encoding_dir: Path = objective_dir.resolve_inside(
                str(encoding(namer)))
            encoding_dir.ensure_dir_exists()
            run_experiment(
                base_dir=encoding_dir,
                instances=inst_creators,
                setups=[
                    cast(Callable, lambda ins, _e=encoding, _o=objective: rls(
                        ins, _e, _o)),
                    cast(Callable, lambda ins, _e=encoding, _o=objective: fea(
                        ins, _e, _o))],
                n_runs=n_runs,
                n_threads=Parallelism.ACCURATE_TIME_MEASUREMENTS,
                perform_warmup=True,
                perform_pre_warmup=True)


# Run the experiment from the command line
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = moptipyapps_argparser(
        __file__, "2D Bin Packing", "Run the 2D Bin Packing experiment.")
    parser.add_argument(
        "dest", help="the directory to store the experimental results under",
        type=Path, nargs="?", default="./results/")
    args: Final[argparse.Namespace] = parser.parse_args()
    run(args.dest)



"""
Different objective functions for two-dimensional bin packing.

The following objective functions are implemented:

- :mod:`~moptipyapps.binpacking2d.objectives.bin_count` returns the number
  of bins occupied by a given packing.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_empty` returns
  a combination of the number of bins occupied by a given packing and the
  fewest number of objects located in any bin.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_last_empty`
  returns a combination of the number of bins occupied by a given packing and
  the number of objects located in the last bin.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_small` returns
  a combination of the number of bins occupied by a given packing and the
  smallest area occupied by objects in any bin.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_last_small` returns
  a combination of the number of bins occupied by a given packing and the
  area occupied by the objects in the last bin.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_lowest_skyline`
  returns a combination of the number of bins occupied by a given packing and
  the smallest area under the skyline in any bin, where the "skyline" is the
  upper border of the space occupied by objects.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_last_skyline`
  returns a combination of the number of bins occupied by a given packing and
  the smallest area under the skyline in the last bin, where the "skyline" is
  the upper border of the space occupied by objects.
"""
